/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * annotate.c
 *
 * Descends recursively into directories and runs a feature creation
 * programme on each detected .jpg file.
 *
 * Stefan M Rueger
 * 
 * 12 April 2005 - Joao Magalhaes:
 * 
 *    - Added single indexing file ('-z'). Index file name is 'index.idx' 
 *      and feature file is '*.ann'.
 *    - Enable/disable recursive functionality ('-r').
 *    - DCT coefficients feature.
 *    - New image rescaling function to generate images with a 
 *      constant number of 8x8 blocks.
 *    - The features filename (function ftfname) uses a "+" instead 
 *      of a ":" for cygwin (Windows) compatibility.
 */

 /*
8th December 2005

Added -y and -w options, this is for specifying the output location
for the index.idx and the .ann file from the -z option

example:
annotate.exe -y /cygdrive/d/Imperial/BT/annotate_new -w /cygdrive/d/Imperial/BT/annotate_new -z -a HDS-184-S

 */



//#ifndef lint
//    static char author[] = "annotate.c 1.6 (c) Copyright 2003 Imperial College London";
//#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include "annotate.h"

void Usage(void);

#define progname "annotate"
#define MAXPATHLEN 1024

Params pars;

Annotations fopts[] = {
    {"thumbnail-<X>x<Y>-(RGB|Y--)", "%dx%d-%3s", "%dx%d-%.3s", "48x48-RGB",
     thumbnail, thumbnailcheck,},

    // "#Tieu and Viola feature vector: 25^3*3 = 46,875 components",
    // "new konvolution added to generate feature without the duplicate elements",
    {"convolution-<L>-(RGB|Ybr|RG-|Y--|...)", "%d-%3s", "%d-%.3s", "3-RGB",
     convolution, convolutioncheck,},
    {"konvolution-<L>-(RGB|Ybr|RG-|Y--|...)", "%d-%3s", "%d-%.3s", "3-RGB",
     convolution, convolutioncheck,},

    // "#joint colour distribution histograms, the three first parameters\n "
    // "indicate subdivisions of the respective axes, the last one (G, L, C or S)\n 

    {"RGB-<R>x<G>x<B>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c", "8x8x8-G",
     rgb,
     XYZcheck,},
    {"HSV_lin-<H>x<S>x<V>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c",
     "8x8x8-G",
     hsvlin, XYZcheck,},
    {"HSV_vol-<H>x<S>x<V>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c",
     "8x8x8-G",
     hsvvol, XYZcheck,},
    {"HSL_lin-<H>x<S>x<L>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c",
     "8x8x8-G",
     hsllin, XYZcheck,},
    {"HSL_vol-<H>x<S>x<L>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c",
     "8x8x8-G",
     hslvol, XYZcheck,},
    {"YpCbCr-<Yp>x<Cb>x<Cr>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c",
     "8x8x8-G",
     ypcbcr, XYZcheck,},
    {"CIELUV-<L>x<U>x<V>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c", "8x8x8-G",
     cieluv, XYZcheck,},
    {"CIELAB-<L>x<A>x<B>-[GFSLC]", "%dx%dx%d-%c", "%dx%dx%d-%c", "8x8x8-G",
     cielab, XYZcheck,},
    {"HDS-184-[GFSLC]", "%3s-%c", "%.3s-%c", "184-G", hds, hdscheck,},

    {"margRGB-<R>+<G>+<B>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", margrgb, margXYZcheck,},
    {"margHSV_lin-<H>+<S>+<V>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", marghsvlin, margXYZcheck,},
    {"margHSV_vol-<H>+<S>+<V>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", marghsvvol, margXYZcheck,},
    {"margHSL_lin-<H>+<S>+<L>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", marghsllin, margXYZcheck,},
    {"margHSL_vol-<H>+<S>+<L>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", marghslvol, margXYZcheck,},
    {"margYpCbCr-<Yp>+<Cb>+<Cr>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c", "10+10+10-H-G", margypcbcr, margXYZcheck,},
    {"margCIELUV-<L>+<U>+<V>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", margcieluv, margXYZcheck,},
    {"margCIELAB-<L>+<A>+<B>-[HM]-[GFSLC]", "%d+%d+%d-%c-%c",
     "%d+%d+%d-%c-%c",
     "10+10+10-H-G", margcielab, margXYZcheck,},

    {"smoothness", "", "", "", smoothness, NULL,},
    {"uniformity", "", "", "", uniformity, NULL,},
    {"variance-<N>-<T>", "%d-%d", "%d-%d", "20-3", variance,
     variancecheck,},

    {"tamuraS-<K>-[ACND]", "%d-%c", "%d-%c", "2-A", tamuraS, tamuraCheck},
    {"tamuraCND-<C>x<N>x<D>-[GFSLC]-<W>-<K>", "%dx%dx%d-%c-%d-%d",
     "%dx%dx%d-%c-%d-%d", "8x8x8-G-8-3", rgb /* sic */ , histTamuraCheck,},
    {"margTamura-<C>+<N>+<D>-[HM]-[GFSLC]-<W>-<K>", "%d+%d+%d-%c-%c-%d-%d",
     "%d+%d+%d-%c-%c-%d-%d", "10+10+10-H-G-8-3", margrgb /* sic */ ,
     margTamuraCheck,},
    {"tamuraImage-<W>-<K>-[ACND]", "%d-%d-%c", "%d-%d-%c", "8-3-A",
     tamuraImage,
     tamuraImageCheck,},

    {"co_occ-[CS]-<Q>-[ENCH]-[AS]", "%c-%d-%c-%c", "%c-%d-%c-%c",
     "C-64-E-A",
     co_occurrence, co_occ_check,},

    {"gabor-<S>-<O>", "%d-%d", "%d-%d", "2-4", gabor, gaborCheck,},
    {"windowgabor-<S>-<O>-<W>-<D>", "%d-%d-%d-%d", "%d-%d-%d-%d",
     "3-4-65-10",
     histGabor, winGaborCheck,},

    {"echo", "", "", "", echo, NULL,},

    {"st-<X>-<bgc>", "%d-%i", "%d-0x%02x", "96-0xee", mksqthumb,
     mksqthumbcheck,},

    //{"dct-<Q>-<N>", "%d-%d", "%d-%d", "4-32", dctCoefs,
    // dctCoefsCheck,},

    //{"wave-<L>-<T>-<P>", "%d-%s-%c", "%d-%s-%c", "3-db1-I", wavelet,
    // waveletCheck,},
};

char **excl;
int nexcl;

char **extargv, *extlabel;
int extargc, extnampos;

static char **torm;
static int ntorm;
void markunlink(char *p)
{
    if (ntorm % 1024 == 0)
        torm = erealloc(torm, (ntorm + 1024) * sizeof *torm);
    torm[ntorm++] = mystrdup(p);
}

void doallunlink()
{
    int i;
    for (i = 0; i < ntorm; i++)
        unlink(torm[i]);
}

void addexcl(char *p)
{
    if (nexcl % 1024 == 0)
        excl = erealloc(excl, (nexcl + 1024) * sizeof *excl);
    excl[nexcl++] = p;
}

int istoexcl(char *p)
{
    int i;
    for (i = 0; i < nexcl; i++)
        if (!strcmp(excl[i], p))
            return 1;
    return 0;
}

int gotexcl(char *fn)
{
    FILE *fp;
    char line[2 * MAXPATHLEN + 1];
    int lno = 0, i;

    errno = 0;
    if ((fp = fopen(fn, "r"))) {
        while (fgets(line, sizeof line, fp)) {
            if ((i = strlen(line)) > 0)
                line[i - 1] = 0;
            addexcl(mystrdup(line));
            lno++;
        }
        fclose(fp);
    }

    if (!lno)
        warning("no or empty %s? Assuming you want to exclude @%s literally.", fn, fn);
    return lno;
}

int numwidth(int n)
{
    if (n > 999)
        return 4;
    if (n > 99)
        return 3;
    if (n > 9)
        return 2;
    return 1;
}

char *ftfname(char *jpgname, int ti, int tj)
{
    static char pams[2048];
    char *p, *ret;

    pams[0] = 0;
    if ((pars.flag.z || pars.flag.p) && pars.ap) {       /* print the parameter string */
        sprintf(pams, "_%s", pars.ap->fname);
        if ((p = index(pams, '-')))
            p[1] = 0;

        if (pars.ap->printline && *pars.ap->printline) {        /* print the parameter string */
            int x[MAXPARAMS];
            int n;

            for (n = 0; n < MAXPARAMS; n++)
                x[n] = pars.what[n] == 'd' || pars.what[n] == 'i' ? *(int *) pars.s[n] : pars.what[n] == 'c' ? *(unsigned char *) pars.s[n] : (int) pars.s[n];
#if MAXPARAMS != 7
            change the line above and make sure in the line below that you have as many x[n] components as MAXPARAMS !
#endif
             sprintf(pams + strlen(pams), pars.ap->printline, x[0], x[1], x[2], x[3], x[4], x[5], x[6]);
        }

        if (pars.maskbottom)
            sprintf(pams + strlen(pams), "+-m%d", pars.maskbottom);
        if (pars.gridG > 0)
            sprintf(pams + strlen(pams), "+G%d-%d-%d", pars.gridG, pars.nblocks, pars.sigma);
        if (pars.gridS > 0)
            sprintf(pams + strlen(pams), "+S%d-%d-%d", pars.gridS, pars.nblocks, pars.sigma);
        if (pars.tx > 1 || pars.ty > 1) {
            if (pars.flag.s)
                sprintf(pams + strlen(pams), "+t%0*dx%0*d", numwidth(pars.tx), ti, numwidth(pars.ty), tj);
            else
                sprintf(pams + strlen(pams), "+T%dx%d", pars.tx, pars.ty);
        }
    }

    ret = strcpy(clrmalloc(strlen(jpgname) + strlen(pams) + strlen(pars.ftfext) + 1), jpgname);
    if ((p = rindex(ret, '.')))
        *p = 0;
    if (*pams)
        strcat(ret, pams);

    if ((strlen(jpgname) != 0) && (pars.flag.z == 0))
        strcat(ret, pars.ftfext);

    return ret;
}

static void parsescanline(Annotations * ap, char *annopt)
{
    char *p = ap->scanline, *q, c, cc;
    int n, ret = 0, nstr = 0, nnum = 0;

    pars.ap = ap;

    /*
     * clear parameters 
     */
    memset(pars.space, 0, sizeof pars.space);
    memset(pars.number, 0, sizeof pars.number);

    while ((c = *p++)) {
        if (c == '%') {
            for (q = p, n = 0; isascii(*p) && isdigit(*p); p++) // optional width
                n = n * 10 + *p - '0';
            switch (cc = *p) {
            case '%':
                p++;
                break;
            case 'i':
            case 'd':
            case 'c':
            case 's':
            case '[':          // ]
                if (ret >= MAXPARAMS)
                    fatal("max %d conversions allowed in %%%s", MAXPARAMS, ap->scanline);
                pars.what[ret] = cc;
                pars.s[ret] = cc == 'd' || cc == 'i' ? (void *) &pars.number[nnum++] : (void *) &pars.space[nstr++][0];
                ret++;
                break;
            default:
                fatal("cannot do this sscan conversion: '%%%s' (unknown conversion char)", q);
            }
            if (p - q > 4 || n > 15 || (cc == 's' && n == 0))
                fatal("cannot cater for this sscan conversion: '%%%s' (max width field is 15)", q);
        }
    }
    /*
     * safeguard 
     */
    for (n = ret; n < MAXPARAMS; n++)
        pars.s[n] = &pars.space[n][0];

#if MAXPARAMS != 7
    please note that after a change of MAXPARAMS you have to change the two sscanf lines below with parameters pars.s[0] ... pars.s[MAXPARAMS - 1]
#endif
    if (ret != sscanf(ap->dfault, ap->scanline, pars.s[0], pars.s[1], pars.s[2], pars.s[3], pars.s[4], pars.s[5], pars.s[6]))
         fatal("cannot initialise scanline (\"%s\") with default parameters (\"%s\")", ap->scanline, ap->dfault);

    if (annopt && *annopt == '-' && ret != sscanf(annopt + 1, ap->scanline, pars.s[0], pars.s[1], pars.s[2], pars.s[3], pars.s[4], pars.s[5], pars.s[6]))
        fatal("cannot read parameters from commandline, was expecting something " "of the form \"%s\", eg, \"...-%s\", but got \"...-%s\".", ap->fname, ap->dfault, annopt + 1);

    while (annopt && (p = index(annopt, ''))) {
        int tx = 1, ty = 1, mnum, gridG, gridS, nblocks, sigma;

        annopt = p + 1;
        switch (p[1]) {
        case 'T':
        case 't':
            if (2 != sscanf(p + 2, "%dx%d", &tx, &ty))
                fatal("was expecting ':T<X>x<Y>', eg :T3x3, dimensions of tiles array, not ':T%s'", p + 2);
            if (tx < 1 || ty < 1)
                fatal(":T%s tiling array dimensions must be at least 1", p + 2);
            if (tx > 1024 || ty > 1024 || tx * ty > 1024)
                fatal("...:T%s: really more than 1024 tiles (need to recompile)?", p + 2);
            pars.tx = tx, pars.ty = ty;
            break;
        case 'G':
        case 'g':
            if (3 != sscanf(p + 2, "%d-%d-%d", &gridG, &nblocks, &sigma))
                fatal("was expecting ':G<gridsz>-<nblocks>-<sigma>', eg :G8-5-50, for smoothing of colour histograms with fixed grid, not ':G%s'", p + 2);
            if (gridG < 1 || nblocks < 1 || sigma < 0)
                fatal(":G%s Gaussian smoothing parameters must be at least 1", p + 2);
            if (nblocks > 32)
                fatal("...:G%s: nblocks (%d) must be in [1,32]", p + 2, nblocks);
            if (gridG * nblocks >= 128)
                fatal("...:G%s: gridsz*nblocks (%d) should be less than 128", p + 2, gridG * nblocks);
            pars.gridG = gridG, pars.nblocks = nblocks, pars.sigma = sigma;
            break;
        case 'S':
        case 's':
            if (3 != sscanf(p + 2, "%d-%d-%d", &gridS, &nblocks, &sigma))
                fatal("was expecting ':S<gridsz>-<nblocks>-<sigma>', eg :S8-5-50, for smoothing of colour histograms with relative grid, not ':S%s'", p + 2);
            if (gridS < 1 || nblocks < 1 || sigma < 0)
                fatal(":S%s Gaussian smoothing parameters must be at least 1", p + 2);
            if (nblocks > 32)
                fatal("...:S%s: nblocks (%d) must be in [1,32]", p + 2, nblocks);
            if (gridS * nblocks >= 128)
                fatal("...:S%s: gridsz*nblocks (%d) should be less than 128", p + 2, gridS * nblocks);
            pars.gridS = gridS, pars.nblocks = nblocks, pars.sigma = sigma;
            break;
        case '-':
            if (1 != sscanf(p + 2, "m%d", &mnum) || mnum < 1 || mnum > 1024)
                fatal("was expecting :-m<num> with <num> between 1 and 1024");
            pars.maskbottom = mnum;
            break;
        default:
            fatal("cannot handle unexpected ':%s' part of -a %s...", p + 1, ap->fname);
        }
    }

    if (pars.gridG && pars.gridS)
        fatal("cannot do both :G<gridsz>-<nblocks>-<sigma> and :S<gridsz>-<nblocks>-<sigma>");

    if (ap->fctcheck && !ap->fctcheck())
        fatal("parameter checking for \"%s\" failed", ap->fname);
}

int main(int argc, char *argv[])
{
    int fargc = 1, opts = 1;
    char *p;
    int i = 0, j, nn;
    int anni = -1;
    char *annopt = NULL, *anntile = NULL;
	char *zfilename, *zfilename2;
    double dbl;
	int pval=0;
	char mychar ='w';

    pars.destsize = 256;
    pars.tx = pars.ty = 1;
    pars.ftfext = ".ftf";
    pars.memleft = 130 * 1024 * 1024;
    while (--argc > 0)
        if (*(p = argv[++i]) != '-' || !opts)
            argv[fargc++] = argv[i];
        else if (opts && !strcmp(p, "--"))
            opts = 0;
        else
            while (*++p)
                switch (*p) {
                case 'h':
                case '?':
                    Usage();
                    exit(0);
                case 'r':
                    pars.flag.r = 1;
                    break;
                case 'R':
                    sscanf(++p,"=%d-%d", &pars.newHeigth, &pars.newWidth);
                    p = "" - 1;
                    break;
                case 'n':
                    pars.flag.n = 1;
                    break;
                case 'U':
                    pars.flag.U++;
                    break;
                case 'C':
                    pars.flag.C = 1;
                    break;
                case 'G':
                    pars.flag.G = 1;
                    break;
                case 'M':
                    if (*++p == '=')
                        p++;
                    dbl = atof(*p ? p : argc > 1 ? (argc--, argv[++i]) : "0");
                    p = "" - 1;
                    if (dbl > 0 && dbl < pow(2., 31.) / 1024 / 1024)
                        pars.memleft = dbl * 1024 * 1024;
                    break;
                case 'f':
                    pars.flag.f++;
                    break;
                case 's':
                    pars.flag.s++;      /* fallthrough: imply -p */
                case 'p':
                    pars.flag.p = 1;
                    break;
                case 'q':
                    pars.flag.q++;
                    break;
                case 'd':
                    pars.flag.d++;
                    break;
				case 'y':
					setname(argv[(i+1)]);
					break;
			/*	case 'w':
					zfilename2 = malloc(500  * sizeof mychar);
					zfilename2 = argv[(i+1)];
					setname2(argv[(i+1)]);
					break;*/
                case 'z':
					pars.flag.z = 1;
                    break;
                case 'x':
                    if (*++p == '=')
                        p++;
                    if ((p = *p ? p : (argc--, argv[++i])))
                        if (*p != '@' || !gotexcl(p + 1))
                            addexcl(p);
                    p = "" - 1;
                    break;
                    /*
                     * case 'i': 
                     */
                    /*
                     * if(*++p=='=') p++; pars.imageext = p; 
                     */
                    /*
                     * p=""-1; break; 
                     */
                case 'k':
                    if (*++p == '=')
                        p++;
                    p = *p ? p : (argc--, argv[++i]);
                    if (p && atoi(p) >= 16 && atoi(p) <= 8192)
                        pars.destsize = atoi(p);
                    else
                        fatal("use -k <number>, number must be between 16 and 8192");
                    p = "" - 1;
                    break;
                case 'm':
                    if (*++p == '=')
                        p++;
                    p = *p ? p : (argc--, argv[++i]);
                    if (p && atoi(p) >= 1 && atoi(p) <= 1024)
                        pars.maskbottom = atoi(p);
                    else
                        fatal("use -m <number>, number must be between 1 and 1024");
                    p = "" - 1;
                    break;

                case 'e':
                    if (*++p == '=')
                        p++;
                    extlabel = p;
                    for (j = i + 1, nn = argc - 1; nn-- > 0; j++)
                        if (!strcmp(argv[j], "{}") && !extnampos)
                            extnampos = j - i - 1;
                        else if (!strcmp(argv[j], ";")) {
                            nn = j - i - 1;
                            break;
                        }
                    if (nn <= 0)
                        fatal("-e needs a command line terminated by a single \\; argument");
                    extargv = clrmalloc((nn + 1) * sizeof *extargv);
                    extargc = nn;
                    for (j = i + 1, nn = 0; nn < extargc; nn++, j++)
                        extargv[nn] = argv[j];
                    i += extargc + 1, argc -= extargc + 1;
                    p = "" - 1;
                    break;

                case 'a':
                    if (*++p == '=')
                        p++;
                    if ((p = *p ? p : (argc--, argv[++i]))) {
                        int k, v = -1, n, l;

                        annopt = index(p, '-');
                        anntile = index(p, '');
                        if (!annopt || (annopt && anntile && anntile < annopt))
                            annopt = anntile;
                        l = annopt ? annopt - p : strlen(p);
                        for (n = k = 0; k < sizeof fopts / sizeof *fopts; k++)
                            if (l > 0 && !strncasecmp(fopts[k].fname, p, l))
                                v = k, n++;
                        if (n != 1)
                            Usage(), fprintf(stderr, "\n"), fatal("-a %s not %s", p, n ? "unique" : "found");
                        anni = v;
                    }
                    p = "" - 1;
                    break;

                default:
                    fprintf(stderr, "unknown option '%c'\n", *p);
                    Usage();
                    exit(1);
                }

    if (extlabel && anni != -1)
        fatal("can do either -e or -a but not both");
    if (anni < 0)
        anni = 0;

    if (fargc == 1) {
        static char *noarg[] = { NULL, ".", NULL };

        fargc = 2;
        noarg[0] = *argv;
        argv = noarg;
    }

    umask(027);
    if (!extlabel)
        parsescanline(&fopts[anni], annopt);
    pars.todo = extlabel ? externalprog : fopts[anni].fct;      /* Function for feature extraction */

    recursive(--fargc, ++argv, ".jpg");

    doallunlink();
    if (pars.flag.d)
        printf("memleft %dMB\n", pars.memleft / 1024 / 1024);
    exit(0);
}


/*
 * VARARGS1 
 */
int warning(fmt, a, b, c, d)
char *fmt;
{
    int err = errno;

    fflush(stdout);
    fprintf(stderr, "%s: ", progname);
    fprintf(stderr, fmt, a, b, c, d);
    if (errno)
        fprintf(stderr, " - "), perror(NULL);
    else
        fprintf(stderr, "\n");
    fflush(stderr);
    errno = 0;

    return err;
}

/*
 * VARARGS1 
 */
void fatal(fmt, a, b, c, d)
char *fmt;
{
    warning(fmt, a, b, c, d);
    exit(1);
}

void Usage(void)
{
    static char *Use[] = {
        "Syntax: ", progname, " {[<file>] [<opts>]}\n",
        "Function: annotates all .jpg files, create annotation files\n",
        "Options:\n",
        "  -n        do not follow softlinks\n",
        "  -x=<file> exclude file/directory (may be repeated)\n",
        "  -U        remove all successfully annotated *.jpg files (except for -a echo)\n",
        "  -UU       remove all found *.jpg files (even for -a echo or when not annotated)\n",
        "  -C        switch off colour histogram index caching\n",
        "  -G        switch off Gaussian smoothing caching (used for some histograms)\n",
        "  -M=<num>  use max <num>MB memory for caching\n",
        "  -q        quiet; don't print the files annotated - just the directories\n",
        "  -qq       very quiet; don't even print the directories\n",
        "  -d        print a few debugging statements\n",
        "  -k=<num>  linearly rescales images to a total of num pixels (num = W*H)\n",
        "  -f        force creation of .ftf files even if they already exist\n",
        "  -p        add parameters to the .ftf filenames, don't use the same basename\n",
        "  -s        splitting: create an ftf file per tile (see below), implies -p\n",
        "  -m=<num>  ignore bottom <num> lines of image before downsampling (TV banner)\n" "  -e=<label> <prog> <args> {} \\;\n",
        "  -z        generate index file and a single annotation file\n",
        "  -r        turn off recursive annotation\n",
        "Annotations:\n",
    };
    static char *Use2[] = {
        "     --\n" "     Note: With all -a annotations one can add :T<X>x<Y> for tiling, ie\n",
        "     the image is dissected into <X> x <Y> nonoverlapping tiles and each\n",
        "     tile is subjected to the feature generation; the overall feature vector\n",
        "     is a concatination of all the feature vectors of the respective tiles.\n\n",
        "     With most histogram annotations one can add :G<gridsz>-<nblocks>-<sigma>\n",
        "     for Gaussian smoothing of the pixels to be entered into the histogram.\n" "     The 256 values per axis are grouped into 3d bins of <gridsz> linear width;\n",
        "     then the <nblocks>**3 nearest bin centres for each pixel are computed\n",
        "     and used with normalised weights proportional to exp(-dist**2/sigma**2)\n",
        "     instead of the pixel coordinates itself.\n",
        "     Alternatively :S<gridsz>-<nblocks>-<sigma> does Gaussian smoothing with a\n",
        "     relative grid moving with the (r,g,b) position of each pixel.\n\n",
        "Example:\n",
        "  annotate -a RGB-8x8x8-G\n",
        "    Recursively descends in the current directory and creates .ftf files, with\n",
        "    the same basename as the .jpg files. These .ftf files contain 8*8*8 floats\n",
        "    which are a global (G) percentage RGB histogram. The file format is machine\n",
        "    dependent; on PCs raw 4-byte float are written with little-endian byte order.\n",
    };
    int i;

    for (i = 0; i < sizeof Use / sizeof *Use; i++)
        fprintf(stderr, "%s", Use[i]);
    for (i = 0; i < sizeof fopts / sizeof *fopts; i++)
        fprintf(stderr, "  -a %s%s\n", fopts[i].fname, i ? "" : " (default)");
    for (i = 0; i < sizeof Use2 / sizeof *Use2; i++)
        fprintf(stderr, "%s", Use2[i]);
}
