/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * variance.c
 * 
 * Author:
 * Peter Howarth, December 2003
 * 
 * Calculates features from co-occurrence matrices
 *
 * 16 distance vectors are used, lengths 1..4 and directions 0, 45, 90
 * and 135 degrees
 * Feature is defined by co_occ-[CS]-<Q>-[ENCH]-[AS]
 * CS: C => 16 matrices created;  S => 4 matrices, summed for the same distances
 * Q:  quantization
 * ENCH: feature calculated, E=energy, N=entropy, C=contrast, H=homogenity
 * AS: asymmetric or symmetrical co-occurance matrix
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"


typedef struct{
  int* mat;
  int size;
  long total;
}Cmatrix;


static void freeCmatrix(Cmatrix);
static Cmatrix newCmatrix(int);
static float energy(Cmatrix);
static float contrast(Cmatrix);
static float entropy(Cmatrix);
static float homogenity(Cmatrix);
static float  (*feature)(Cmatrix);

// INPUTS 
static int quant, symm, cat;
static int rads = 4;
static int vectors[] = {1, 0,  1, 1,  0, 1,  -1, 1};
static int no_of_vectors = 4;
static int no_of_matrices = 0;
/*
 * Creates a co-occ matrix for each combination of vector (direction) and radius
 * and then calculates a feature for each matrix. The default inputs have 4 vectors
 * and distances resulting in a feature vector of 16 floats
 */ 
int co_occurrence(FILE *fp, AImage im) {
  int i, j, v, r, x, y, z, a, b, m;
  float result[no_of_matrices];
  Cmatrix cm[no_of_matrices];
  AImage g_im = ConvertToGrey(im, quant);
  if (!no_of_matrices)
    fatal("no_of_matrices is zero");
  for (z=0; z < no_of_matrices; z++){
    cm[z] = newCmatrix(quant);
  }

  for (i=0; i < g_im.h; i++) {
    for (j=0; j < g_im.w; j++) {
      // update each co-occ matrix
      for (r = 1; r <= rads; r++) {
	for (v = 0; v < no_of_vectors; v++) {
	  m = cat ? (r-1)*no_of_vectors + v: (r-1); 
	  x = i + vectors[2*v] * r;
	  y = j + vectors[2*v + 1] * r;
	  if (x>=0 && y>=0 && x<g_im.h && y<g_im.w) {
	    a = (int)g_im.cont[x*g_im.w + y];
	    b = (int)g_im.cont[i*g_im.w + j];
	    cm[m].mat[a*quant + b]++;
	    cm[m].total++;
	  }
	  if(symm) {
	    x = i - vectors[2*v] * r;
	    y = j - vectors[2*v + 1] * r;
	    if (x>=0 && y>=0 && x<g_im.h && y<g_im.w) {
	      a = (int)g_im.cont[x*g_im.w + y];
	      b = (int)g_im.cont[i*g_im.w + j];
	      cm[m].mat[a*quant + b]++;
	      cm[m].total++;
	    }
	  }
	}
      }
    }
  }
   for (z=0; z < no_of_matrices; z++) { 
    if(!feature)
      fatal("how come feature() was not set?");
    result[z] = feature(cm[z]);
    if (DEBUG)
      printf("feature=%.5f\n",  feature(cm[z]));
    freeCmatrix(cm[z]);
  }
  fwrite(result, sizeof*result, sizeof result/sizeof*result, fp);
  efree(g_im.cont);
  return 1;
}

int co_occ_check() {
  quant = pars.number[0];
  if (quant < 2 || quant > 256) {
    fatal("Quantization should be between 2 and 256.  Default value is 64");
  }
  switch(*pars.space[0]) {
  case 'C': case 'c': cat = 1; no_of_matrices = no_of_vectors * rads; break;
  case 'S': case 's': cat = 0; no_of_matrices = no_of_vectors; break;
  default:
    fatal("must specify C or S in co_occ-[CS]-<x>-[x]-[x]. C: separate co-occ matrices for each distance vector - S: sum the same length distance vectors (4 directions) into one co-occ matrix");
  }

  switch(*pars.space[1]) {
  case 'E': case 'e': *pars.space[1]='E'; feature = energy;     break;
  case 'N': case 'n': *pars.space[1]='N'; feature = entropy;    break;
  case 'C': case 'c': *pars.space[1]='C'; feature = contrast;   break;
  case 'H': case 'h': *pars.space[1]='H'; feature = homogenity; break;
  default:
    fatal("must specify E, N, C or H in co_occ-[x]-<x>-[ENCH]-[x]. E, N, C, H represent the features for energy, entropy, contrast and homogenity.");
  }
  switch(*pars.space[2]) {
  case 'A': case 'a': symm = 0; break;
  case 'S': case 's': symm = 1; break;
  default:
    fatal("must specify A or S in co_occ-[x]-<x>-[x]-[AS]. A and S represent assymetric and symetric matrices");
  }
  return 1;
}


float energy(Cmatrix cm) {
  int i;
  double value, energy = 0;
  for (i=0; i< cm.size * cm.size; i++) {
    value =  (double)cm.mat[i]/(double)cm.total;  
    energy += pow(value, 2);
  }
  return (float)energy;
}

float contrast(Cmatrix cm) {
  int i, j;
  double value, contrast = 0;
  for (i=0; i<cm.size; i++) {
    for (j=0; j<cm.size; j++) {
      value =  (double)cm.mat[i*cm.size + j]/(double)cm.total;
      contrast += (pow(i-j, 2) * value);
    }
  }
  return (float)contrast;
}
float entropy(Cmatrix cm) {
  int i, j;
  double value, entropy = 0;
  for (i=0; i<cm.size; i++) {
    for (j=0; j<cm.size; j++) {
      value =  (double)cm.mat[i*cm.size + j]/(double)cm.total;
      if (value)
	entropy -= value * log(value);
    }
  }
  return (float)entropy;
}
float homogenity(Cmatrix cm) {
  int i, j;
  double value, homogenity = 0;
  for (i=0; i<cm.size; i++) {
    for (j=0; j<cm.size; j++) {
      value =  (double)cm.mat[i*cm.size + j]/(double)cm.total;
      homogenity += (value / (1 + abs(i-j)));
    }
  }
  return (float)homogenity;
}

Cmatrix newCmatrix(int size) {
  Cmatrix ret;
  ret.size = size;
  ret.total = 0;
  ret.mat = clrmalloc(size * size * sizeof(int));
  return ret;
}

void freeCmatrix(Cmatrix cm) {
  efree(cm.mat);
}
