/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * colourhist.c
 *
 * Creates various 3dim colour histograms from an image.
 *
 * Stefan M Rueger
 *
 * For each colour model we support several major methods:
 *
 *
 * G (for global)
 * computes a global percentage colour histogram
 * There are XXXNcells(N1,N2,N3) 3-d colour bins and consequently
 * the same number of float numbers which make up the feature vector.
 * The sum of the components is 1.
 *
 *
 * F (for focus)
 * computes a global percentage colour histogram in the 25% centre area of 
 * the image. There are XXXNcells(N1,N2,N3) 3-d colour bins and, consequently,
 * the same number of float numbers which make up the feature vector.
 * The sum of the components is 1.
 *
 *
 * S (for structured)
 * computes a structured colour histogram
 * For each pixel in the image the correspeonding bin is computed.
 * A gliding 8x8 window  moves over the image - each histogram
 * bin counts the number of windows which contain at least one
 * pixel of the corresponding bin. Local information is folded
 * into this type of histogram. Normalisation is through dividing
 * by the number of different window positions; each component
 * is thus between 0 and 1, but the sum can be as large as 64 (in rare
 * cases).
 *
 *
 * C (for centre)  .-------------------.
 *                 |  0      |      2  |
 *                 |   .-----------.   |
 *                 |   |           |   |
 *                 |---|     4     |---|
 *                 |   |           |   |
 *                 |   .-----------.   |
 *                 |  1      |      3  |
 *                 .-------------------.
 * splits the image into 5 areas: centre (50% area) and 4 quarters
 * of the border. For each area a percentage histogram for is computed:
 * For example in RGB there are NR*NG*NB 3-d colour bins and consequently
 * 5*NR*NG*NB float numbers which make up the feature vector.
 * The central histogram is weighted with a relative factor of 16, so
 * that (considering each of the 4 border tiles weighs 1) its contribution
 * is 80% of the image. The four border tiles have a cumulative importance 
 * of 20%. The sum of all the components is made to be 1.
 *
 * 
 * L (for local)   .-------------------.
 *                 |                   |
 *                 |   .-----------.   |
 *                 |   |  0  |  2  |   |
 *                 |   |-----+-----|   |
 *                 |   |  1  |  3  |   |
 *                 |   .-----------.   |
 *                 |                 4 |
 *                 .-------------------.
 * splits the image into 5 areas: border (50% area) and 4 quarters
 * For each area a percentage histogram for is computed:
 * For example in RGB there are NR*NG*NB 3-d colour bins and consequently
 * 5*NR*NG*NB float numbers which make up the feature vector.
 * As each area histogram is normalised separately, the middle
 * histograms are weighed higher than the border; they contribute
 * 20% each to the feature vector (but they have only 12.5% area)
 * The border is demoted from 50% area to only 20% contribution.
 * The sum of all the components is made to be 1.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>

#include "annotate.h"
#include "colourmodels.c"

static struct {
    int hsize;
    int (*cache) (int, int, int);
    int (*cell) (int, int, int);
    Coord *(*smoothing) (int, int, int);
} args;

static unsigned char *uc_m;
static unsigned short *us_m;
static int *ul_m;
#define CACHESIZ (256*256*256)

static int uc_cache(int r, int g, int b)
{
    int ind = (((r << 8) + g) << 8) + b, ret = uc_m[ind];
    if (ret == 255)
        uc_m[ind] = ret = args.cell(r, g, b);
    return ret;
}

static int us_cache(int r, int g, int b)
{
    int ind = (((r << 8) + g) << 8) + b, ret = us_m[ind];
    if (ret == 65535)
        us_m[ind] = ret = args.cell(r, g, b);
    return ret;
}

static int ul_cache(int r, int g, int b)
{
    int ind = (((r << 8) + g) << 8) + b, ret = ul_m[ind];
    if (ret == -1L)
        ul_m[ind] = ret = args.cell(r, g, b);
    return ret;
}


static int
globalcolourhist(FILE * fp, AImage im),
focuscolourhist(FILE * fp, AImage im),
structuredcolourhist(FILE * fp, AImage im), centrecolourhist(FILE * fp, AImage im), localcolourhist(FILE * fp, AImage im), (*torun) (FILE * fp, AImage im);

static int runhistogram(FILE * fp, AImage im, int (*Ncells) (int, int, int), int (*cell) (int, int, int))
{
    args.smoothing = pars.gridS ? smoothing_S : pars.gridG ? smoothing_G : NULL;
    if (!args.cache) {
        int size;
        args.cache = args.cell = cell;
        args.hsize = Ncells(pars.number[0], pars.number[1], pars.number[2]);
        if (!pars.flag.C) {
            if (args.hsize < 255) {
                if ((size = CACHESIZ * sizeof *uc_m) < pars.memleft && (uc_m = malloc(size))) {
                    memset(uc_m, -1, size), args.cache = uc_cache, pars.memleft -= size;        // printf("uc cache\n");
                    if (pars.flag.d)
                        printf("cache 1\n");
                }
            } else if (args.hsize < 65535) {
                if ((size = CACHESIZ * sizeof *us_m) < pars.memleft && (us_m = malloc(size))) {
                    memset(us_m, -1, size), args.cache = us_cache, pars.memleft -= size;        // printf("us cache\n");
                    if (pars.flag.d)
                        printf("cache 2\n");
                }
            } else {
                if ((size = CACHESIZ * sizeof *ul_m) < pars.memleft && (ul_m = malloc(size))) {
                    memset(ul_m, -1, size), args.cache = ul_cache, pars.memleft -= size;        // printf("ul cache\n");
                    if (pars.flag.d)
                        printf("cache 4\n");
                }
            }
        }
    }
    if (!torun)
        fatal("how come torun() was not set?");

    return torun(fp, im);
}


#define NT 5                    /* tiles */
#define BF 0.14644660940673     /* ((1-sqrt(.5))/2) such that border = 50%  */

static int localcolourhist(FILE * fp, AImage im)
{
    int i, j, k, ci, r, g, b;
    float *c_hist, *c_p;
    double n_hist;
    unsigned char *c;
    Coord *co;
    int xl = BF * im.w + 1, xr = (1 - BF) * im.w, yt = BF * im.h + 1, yb = (1 - BF) * im.h, xm = im.w / 2, ym = im.h / 2;

    c_hist = clrmalloc(NT * args.hsize * sizeof *c_hist);

    if (args.smoothing) {       // smooth each colour pixel with a gaussian function
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3) {
                k = i < xl || i > xr || j < yt || j > yb ? /* border */ 4 : /* 4 inner */ (i > xm) * 2 + (j > ym);
                co = args.smoothing(c[0], c[1], c[2]);
                for (r = co->r, g = co->g, b = co->b, ci = 0; ci < co->N; ci++)
                    c_hist[k * args.hsize + args.cache(r + co->c[ci].dr, g + co->c[ci].dg, b + co->c[ci].db)]
                        += co->c[ci].w;
            }
    } else
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3) {
                k = i < xl || i > xr || j < yt || j > yb ? /* border */ 4 : /* 4 inner */ (i > xm) * 2 + (j > ym);
                c_hist[k * args.hsize + args.cache(c[0], c[1], c[2])]++;
            }

    for (k = 0; k < NT; k++) {
        for (n_hist = 0, c_p = c_hist + k * args.hsize, i = 0; i < args.hsize; i++)
            n_hist += c_p[i];
        if (n_hist *= 5)
            for (i = 0; i < args.hsize; i++)
                c_p[i] /= n_hist;
    }

    fwrite(c_hist, sizeof *c_hist, NT * args.hsize, fp);
    efree(c_hist);

    return 1;
}


static int centrecolourhist(FILE * fp, AImage im)
{
    int i, j, k, ci, r, g, b;
    float *c_hist, *c_p;
    double n_hist;
    unsigned char *c;
    Coord *co;
    int xl = BF * im.w + 1, xr = (1 - BF) * im.w, yt = BF * im.h + 1, yb = (1 - BF) * im.h, xm = im.w / 2, ym = im.h / 2;

    c_hist = clrmalloc(NT * args.hsize * sizeof *c_hist);

    if (args.smoothing) {       // smooth each colour pixel with a gaussian function
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3) {
                k = i >= xl && i <= xr && j >= yt && j <= yb ? /* centre */ 4 : /* 4 outer */ (i > xm) * 2 + (j > ym);
                co = args.smoothing(c[0], c[1], c[2]);
                for (r = co->r, g = co->g, b = co->b, ci = 0; ci < co->N; ci++)
                    c_hist[k * args.hsize + args.cache(r + co->c[ci].dr, g + co->c[ci].dg, b + co->c[ci].db)]
                        += k == 4 ? 16 * co->c[ci].w : co->c[ci].w;
            }
    } else
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3) {
                k = i >= xl && i <= xr && j >= yt && j <= yb ? /* centre */ 4 : /* 4 outer */ (i > xm) * 2 + (j > ym);
                c_hist[k * args.hsize + args.cache(c[0], c[1], c[2])] += k == 4 ? 16 : 1;
            }

    for (k = 0; k < NT; k++) {  // normalise histograms
        for (n_hist = 0, c_p = c_hist + k * args.hsize, i = 0; i < args.hsize; i++)
            n_hist += c_p[i];
        n_hist = k == 4 ? 20 * n_hist / 16 : 20 * n_hist;
        if (n_hist)
            for (i = 0; i < args.hsize; i++)
                c_p[i] /= n_hist;
    }

    fwrite(c_hist, sizeof *c_hist, NT * args.hsize, fp);
    efree(c_hist);

    return 1;
}


static int globalcolourhist(FILE * fp, AImage im)
{
    int i, j, ci, r, g, b;
    float *c_hist;
    double n_hist;
    unsigned char *c;
    Coord *co;

    c_hist = clrmalloc(args.hsize * sizeof *c_hist);

    if (args.smoothing) {       // smooth each colour pixel with a gaussian function
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3) {
                // double sum=0; printf("\n-----------\n%3d %3d %3d\n-----------\n", c[0], c[1], c[2]);
                co = args.smoothing(c[0], c[1], c[2]);
                for (r = co->r, g = co->g, b = co->b, ci = 0; ci < co->N; ci++) {
                    // printf("%3d %3d %3d: %4.2f\n", r+co->c[ci].dr, g+co->c[ci].dg, b+co->c[ci].db, co->c[ci].w*100);
                    c_hist[args.cache(r + co->c[ci].dr, g + co->c[ci].dg, b + co->c[ci].db)] += co->c[ci].w;
                    // sum+=co->c[ci].w;
                }
                // printf("sum = %4.2f\n", 100*sum);
            }
    } else
        for (c = im.cont, j = 0; j < im.h; j++)
            for (i = 0; i < im.w; i++, c += 3)
                c_hist[args.cache(c[0], c[1], c[2])]++;

    for (n_hist = 0, i = 0; i < args.hsize; i++)        // in theory n_hist == im.w*im.h except for rounding errors
        n_hist += c_hist[i];
    if (n_hist)
        for (i = 0; i < args.hsize; i++)
            c_hist[i] /= n_hist;

    fwrite(c_hist, sizeof *c_hist, args.hsize, fp);
    efree(c_hist);

    return 1;
}


static int focuscolourhist(FILE * fp, AImage im)
{
    int i, j, ci, r, g, b, xl, xr, yt, yb;
    float *c_hist;
    double n_hist;
    unsigned char *c;
    Coord *co;

    c_hist = clrmalloc(args.hsize * sizeof *c_hist);

    xl = .25 * im.w, xr = 0.75 * im.w, yt = 0.25 * im.h, yb = 0.75 * im.h;
    if ((xr - xl) > 0 && (yb - yt) > 0) {
        if (args.smoothing) {   // smooth each colour pixel with a gaussian function
            for (j = yt; j < yb; j++)
                for (i = xl, c = im.cont + (j * im.w + i) * 3; i < xr; i++, c += 3) {
                    co = args.smoothing(c[0], c[1], c[2]);
                    for (r = co->r, g = co->g, b = co->b, ci = 0; ci < co->N; ci++)
                        c_hist[args.cache(r + co->c[ci].dr, g + co->c[ci].dg, b + co->c[ci].db)] += co->c[ci].w;
                }
        } else
            for (j = yt; j < yb; j++)
                for (i = xl, c = im.cont + (j * im.w + i) * 3; i < xr; i++, c += 3)
                    c_hist[args.cache(c[0], c[1], c[2])]++;
    }

    for (n_hist = 0, i = 0; i < args.hsize; i++)        // n_hist should be (xr-xl)*(yb-yt) save for rounding errors
        n_hist += c_hist[i];
    if (n_hist)
        for (i = 0; i < args.hsize; i++)
            c_hist[i] /= n_hist;

    fwrite(c_hist, sizeof *c_hist, args.hsize, fp);
    efree(c_hist);

    return 1;
}


static int structuredcolourhist(FILE * fp, AImage im)
{
    int i, j, ci, cj, k, n_hist, *c_bin, bin;
    float *c_hist;
    unsigned char *c, *c_mark;

    if (pars.gridS || pars.gridG)
        fatal("cannot use -S structured colour histogram with :S<gridsz>-... or :G<gridsz>-... smoothing modifier");

    c_hist = clrmalloc(args.hsize * sizeof *c_hist);
    c_mark = clrmalloc(args.hsize * sizeof *c_mark);

    c_bin = clrmalloc(im.h * im.w * sizeof *c_bin);
    for (c = im.cont, k = j = 0; j < im.h; j++)
        for (i = 0; i < im.w; i++, c += 3)
            c_bin[k++] = args.cache(c[0], c[1], c[2]);

    for (j = 0; j < im.h - 7; j++)
        for (i = 0; i < im.w - 7; i++)
            for (memset(c_mark, 0, args.hsize * sizeof *c_mark), cj = 0; cj < 8; cj++)
                for (ci = 0; ci < 8; ci++)
                    if (!c_mark[bin = c_bin[(j + cj) * im.w + i + ci]])
                        c_mark[bin] = 1, c_hist[bin]++;

    efree(c_bin);
    efree(c_mark);

    if ((n_hist = (im.h - 7) * (im.w - 7)) > 0)
        for (i = 0; i < args.hsize; i++)
            c_hist[i] /= n_hist;

    fwrite(c_hist, sizeof *c_hist, args.hsize, fp);
    efree(c_hist);

    return 1;
}


int XYZcheck()
{
    int i;
    for (i = 0; i < 3; i++)
        if (pars.number[i] <= 0 || pars.number[i] > 256)
            fatal("N%c-axis subdivision (%d) in <colourmodel>-N1xN2xN3-[GCLS] must be in range of [1,256]", '1' + i, pars.number[i]);

    switch (*pars.space[0]) {
    case 'G':
    case 'g':
        *pars.space[0] = 'G';
        torun = globalcolourhist;
        break;
    case 'F':
    case 'f':
        *pars.space[0] = 'F';
        torun = focuscolourhist;
        break;
    case 'S':
    case 's':
        *pars.space[0] = 'S';
        torun = structuredcolourhist;
        break;
    case 'C':
    case 'c':
        *pars.space[0] = 'C';
        torun = centrecolourhist;
        break;
    case 'L':
    case 'l':
        *pars.space[0] = 'L';
        torun = localcolourhist;
        break;
    default:
        fatal("must specify G, F, S, C or L in <colourmodel>-N1xN2xN3-[GFSCL]. G, F, S, C, or L stand for global, focus, structured, centred or local histogram types");
    }
    return 1;
}

int rgb(FILE * fp, AImage im)
{
    pars.flag.C = 1;            /* no caching here - easy to cmopute */
    return runhistogram(fp, im, RGBNcells, RGBcell);
}

int hsvlin(FILE * fp, AImage im)
{
    return runhistogram(fp, im, HSVNcells, HSVlincell);
}

int hsvvol(FILE * fp, AImage im)
{
    return runhistogram(fp, im, HSVNcells, HSVvolcell);
}

int hsllin(FILE * fp, AImage im)
{
    return runhistogram(fp, im, HSLNcells, HSLlincell);
}

int hslvol(FILE * fp, AImage im)
{
    return runhistogram(fp, im, HSLNcells, HSLvolcell);
}

int cieluv(FILE * fp, AImage im)
{
    return runhistogram(fp, im, CIELUVNcells, CIELUVcell);
}

int cielab(FILE * fp, AImage im)
{
    return runhistogram(fp, im, CIELABNcells, CIELABcell);
}

int ypcbcr(FILE * fp, AImage im)
{
    return runhistogram(fp, im, YpCbCrNcells, YpCbCrcell);
}

int hdscheck()
{
    if (strcmp("184", pars.space[0]))
        fatal("can only do HDS-184, not HDS-%s", pars.space[0]);
    switch (*pars.space[1]) {
    case 'G':
    case 'g':
        *pars.space[1] = 'G';
        torun = globalcolourhist;
        break;
    case 'F':
    case 'f':
        *pars.space[1] = 'F';
        torun = focuscolourhist;
        break;
    case 'S':
    case 's':
        *pars.space[1] = 'S';
        torun = structuredcolourhist;
        break;
    case 'C':
    case 'c':
        *pars.space[1] = 'C';
        torun = centrecolourhist;
        break;
    case 'L':
    case 'l':
        *pars.space[1] = 'L';
        torun = localcolourhist;
        break;
    default:
        fatal("must specify G, F, S, C or L in HDS-184-[GFSCL]. G, F, S, C, or L stand for global, focus, structured, centred or local histogram types");
    }
    return 1;
}

int hds(FILE * fp, AImage im)
{
    return runhistogram(fp, im, HDSNcells, HDScell);
}
