/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * colour models
 *
 * Stefan M Rueger
 *
 * For each colour model we define two functions
 *
 *   <colourmodel>Ncells(int N1, int N2, int N3)
 * 
 * and
 *
 *   <colourmodel>cell(int r, int g, int b)
 *
 * The Ncells() call sets up the dimensions of the bins and computes
 * the number of bins; each of the following cell(r,g,b) calls returns
 * an index between 0 and the number of bins -1. R, g, and b must be in
 * the range of [0,255].
 *
 * We assume that each colourmodel we deal with has 3 independent
 * coordinates which are to be split up into N1, N2 and N3 bins.
 * Ordinarily, one would think there are N1*N2*N3 different bins by
 * splitting up the colourspace. However, some colourmodels such as HSL
 * have a periodic coordinate such as Hue, and are best represented as
 * cylindrical coordinates: In the case of HSL, a low S value makes the
 * colour lose it's  chroma, and hue is undefined (or irrelevant). As a
 * consequence, all (H,S=0) values are identified into one cell, and a
 * slightly smaller number of cells than N1*N2*N3 is assumed.
 * Also note that the colour histograms deducted from MPEG-7 prescribe
 * the splitting up of the colour space - N1, N2 and N3 are not used here.
 *
 */


static int n1, n2, n3, N1, N2, N3;

/*
 * RGB space is the most simple
 *
 */

static int RGBNcells(int NR, int NG, int NB) {
  N1=NR, N2=NG, N3=NB; 
  return N1*N2*N3; 
}

static int RGBcell(int r, int g, int b) {
  return (n1 = N1*r/256)*N2*N3 + (n2 = N2*g/256)*N3 + (n3 = N3*b/256);
}


/*
 * Y'CbCr is a linear transformation from gamma-corrected (R',G',B')
 * values. The coefficients used here are for Recommendation ITU-R
 * BT.601-4,  Recommendation ITU-R BT.470-4 System B and G, SMPTE
 * 170M-1994 and SMPTE 293M-1996. 
 */

static int YpCbCrNcells(int NYp, int NCb, int NCr) { 
  N1=NYp, N2=NCb, N3=NCr;
  return N1*N2*N3; 
}

static int YpCbCrcell(int r, int g, int b) {
  double Yp, Cb, Cr, fr = 0.299, fg = 0.587, fb = 0.114;
  int yp, cb, cr;

  Yp =           fr*r/255. +          fg*g/255. +          fb*b/255.;
  Cb = -fr/2/(1-fb)*r/255. - fg/2/(1-fb)*g/255. +         0.5*b/255.;
  Cr =          0.5*r/255. - fg/2/(1-fr)*g/255. - fb/2/(1-fr)*b/255.;

  n1 = yp =       Yp*N1*.99999999;
  n2 = cb = (0.5+Cb)*N2*.99999999;
  n3 = cr = (0.5+Cr)*N3*.99999999;

  if(yp<0 || cb<0 || cr<0 || yp>=N1 || cb>=N2 || cr>=N3) {
    printf("%6g %6g %6g %d %d %d\n", Yp, Cb, Cr, yp, cb, cr);
    fatal("Y'CbCr conversion: vales out of range?");
  }

  return yp*N2*N3 + cb*N3 + cr;
}


/*
 * CIELUV and CIELAB space, more perceptually uniform
 *
 *  see also http://www.cs.rit.edu/~ncs/color/ and click on t_convert.html
 * 
 *  [ X ]   [  0.412453  0.357580  0.180423 ]   [ R ]
 *  [ Y ] = [  0.212671  0.715160  0.072169 ] * [ G ]
 *  [ Z ]   [  0.019334  0.119193  0.950227 ]   [ B ]
 * 
 * 
 * 
 * L* = 116 * (Y/Yn)^(1/3) - 16    for Y/Yn > 0.008856
 * L* = 903.3 * Y/Yn               otherwise
 * 
 * a* = 500 * ( f(X/Xn) - f(Y/Yn) )
 * b* = 200 * ( f(Y/Yn) - f(Z/Zn) )
 *     where f(t) = t^(1/3)      for t > 0.008856
 *               f(t) = 7.787 * t + 16/116    otherwise
 * 
 * 
 * ---
 * 
 * L* =  116 * (Y/Yn)1/3 - 16
 * u* =  13L* * ( u' - un' )
 * v* =  13L* * ( v' - vn' )
 * 
 * The quantities un' and vn'  refer to the reference white or the light
 * source; for the 20 observer and illuminant C,  un' = 0.2009, vn' = 0.4610
 * [1]. Equations for u' and v' are given below:
 * 
 *         u' = 4X / (X + 15Y + 3Z) = 4x / ( -2x + 12y + 3 )
 *         v' = 9Y / (X + 15Y + 3Z) = 9y / ( -2x + 12y + 3 ).
 * 
 */


/* not all cells are populated though */
static int CIELABNcells(int NL, int NA, int NB) { 
  N1=NL, N2=NA, N3=NB;
  return N1*N2*N3; 
}

static int CIELABcell(int r, int g, int b) {
  double XbXn, YbYn, ZbZn, Ls, as, bs; 
  int nl, na, nb;

#define WR 255
#define WG 255
#define WB 255

#define Xn (0.412453*WR + 0.357580*WG + 0.180423*WB)
#define Yn (0.212671*WR + 0.715160*WG + 0.072169*WB)
#define Zn (0.019334*WR + 0.119193*WG + 0.950227*WB)

  XbXn = (0.412453*r + 0.357580*g + 0.180423*b)/Xn;
  YbYn = (0.212671*r + 0.715160*g + 0.072169*b)/Yn;
  ZbZn = (0.019334*r + 0.119193*g + 0.950227*b)/Zn;

#define fscale(x) ((x)>0.008856? pow((x),1/3.): 7.787*(x) + 16/116.)
  Ls = 116*fscale(YbYn) - 16;            /* range [0, 100] */
  as = 500*(fscale(XbXn)-fscale(YbYn));  /* range [-86.1813, 98.2352] */
  bs = 200*(fscale(YbYn)-fscale(ZbZn));  /* range [-107.862, 94.4758]*/

  n1 = nl = (Ls-0)/(100-0)*N1*.99999999;
  n2 = na = (as+86.1814)/(98.2353+86.1814)*N2;
  n3 = nb = (bs+107.863)/(94.4759+107.863)*N3;
  if(nl<0 || na<0 || nb<0 || nl>=N1 || na>=N2 || nb>=N3) {
    printf("%6g %6g %6g %d %d %d\n", Ls, as, bs, nl, na, nb);
    fatal("CIE L*a*b* conversion: vales out of range?");
  }

  return nl*N2*N3 + na*N3 + nb;
}


static int CIELUVNcells(int NL, int NU, int NV) {
  N1=NL, N2=NU, N3=NV;
  return N1*N2*N3; 
}

static int CIELUVcell(int r, int g, int b) {
  double X, Y, Z, Ls, us, vs; 
  int nl, nu, nv;

  X = 0.412453*r + 0.357580*g + 0.180423*b;
  Y = 0.212671*r + 0.715160*g + 0.072169*b+0.0000001;
  Z = 0.019334*r + 0.119193*g + 0.950227*b;

#define up(X,Y) (4*(X)/((X)+15*(Y)+3*(Z)))
#define vp(X,Y) (9*(Y)/((X)+15*(Y)+3*(Z)))

  Ls = 116*fscale(Y/Yn) - 16;          /* range [0, 100] */
  us = 13*Ls*(up(X,Y) - up(Xn,Yn));  /* range [-123.324, 147.573] */
  vs = 13*Ls*(vp(X,Y) - vp(Xn,Yn));  /* range [-139.267, 12.1182] */

  n1 = nl = (Ls-0)/(100-0)*N1*.99999999;
  n2 = nu = (us+123.325)/(147.574+123.325)*N2;
  n3 = nv = (vs+139.268)/(12.1183+139.268)*N3;
  if(nl<0 || nu<0 || nv<0 || nl>=N1 || nu>=N2 || nv>=N3) {
    printf("%6g %6g %6g %d %d %d\n", Ls, us, vs, nl, nu, nv);
    fatal("CIE L*u*v* conversion: vales out of range?");
  }

  return nl*N2*N3 + nu*N3 + nv;
}


/*
 * HSV space (= cone) and HSL space (= double cone)
 *
 * H = Hue, S= saturation, V = brightness computed as max(r,g,b)
 *
 * As a cosequence, there are quite a few (r,g,b) pixels with V=1.
 * but only one with V=0, this is often visualised as a cone with the
 * apex symbolising V=0 and the base for V=1. Mathematically, HSV space
 * is a cylinder, though with H in [0,360) degrees, and L, S in [0,1].
 *
 * H = hue, S = saturation, L = brightness as (max(r,g,b)-min(r,g,b))/2
 *
 * There is one (r,g,b) pixel with L=0 and one with L=1. Most (r,g,b) pixels
 * mapped onto a particular L value will be for L=0.5. This is referred to
 * double-cone, with two apices at L=0 and L=1, respectively. Mathematically,
 * HSL space is a cylinder with H and S being radial coordinates and L
 * being the height.
 * 
 */

static int HSVNcells(int NH, int NS, int NV) {
  N1=NH, N2=NS, N3=NV;
  return ((NV)*(1+((NS)-1)*(NH)));
}

/*
 *
 * HSXcelllinear cuts up the volume of all possible H, S and X = L or V
 * values  in a linear way, ie, H space is linearly divided up in NH 
 * cells, S space is linearly divided up in NS cells and X space is linearly
 * cut up in NX cells. But note that all H cells for which S=0 are merged
 * into one bin.
 *
 * HSVcellvolume cuts up the volume of all possible H, S and V values into
 * the same number of cells as with above method, but seeks a non-linear
 * division of each axis, so that the resuting bins cover all the same 
 * number of (r,g,b) pixels of the whole RGB cube.
 *
 * HSLcellvolume cuts up the volume of all possible H, S and L values into
 * the same number of cells as with above method, but seeks a non-linear
 * division of each axis, so that the resuting bins cover all the same 
 * number of (r,g,b) pixels of the whole RGB cube.
 *
 *
 */

static int HSXcelllinear(double H, double S, double X) {
  int h, s, x;

  n1 = h = H/360*N1;
  n2 = s = S*N2*.99999999;
  n3 = x = X*N3*.99999999;
  if(h<0 || s<0 || x<0 || h>=N1 || s>=N2 || x>=N3) {
    printf("%6g %6g %6g %d %d %d\n", H, S, X, h, s, x);
    fatal("HSX linear conversion: vales out of range?");
  }

  /* conflate all hue cells where s=0 (grey, h undefined or nearly irrelevant) */

  return x*(1+(N2-1)*N1) + (s? 1+(s-1)*N1+h: 0);
}

static int HSVcellvolume(double H, double S, double V) {
  int h, s, v;

  n1 = h = H/360*N1;
  n2 = s = (1+(S*S*(N1*(N2-1)+1)-1)/N1)*.99999999;
  n3 = v = V*V*N3*.99999999;
  if(h<0 || s<0 || v<0 || h>=N1 || s>=N2 || v>=N3) {
    printf("%6g %6g %6g %d %d %d\n", H, S, V, h, s, v);
    fatal("HSV volume conversion: vales out of range?");
  }

  /* conflate all hue cells where s=0 (grey, h undefined or nearly irrelevant) */

  return v*(1+(N2-1)*N1) + (s? 1+(s-1)*N1+h: 0);
}


static int HSLcellvolume(double H, double S, double L) {
  int h, s, l;

  n1 = h = H/360*N1;
  n2 = s = (1+(S*S*(N1*(N2-1)+1)-1)/N1)*.99999999;
  n3 = l = (L<.5? 2*L*L: 1-2*(1-L)*(1-L))*N3*.99999999;
  if(h<0 || s<0 || l<0 || h>=N1 || s>=N2 || l>=N3) {
    printf("%6g %6g %6g %d %d %d\n", H, S, L, h, s, l);
    fatal("HSL volume conversion: vales out of range?");
  }

  /* conflate all hue cells where s=0 (grey, h undefined or nearly irrelevant) */

  return l*(1+(N2-1)*N1) + (s? 1+(s-1)*N1+h: 0);
}


static int allHSVcell(int r, int g, int b, int (*HSXcell)(double H, double S, double V)) {
  int min, max;
  double delta, H, S, V;

  min=r; if(g<min) min=g; if(b<min) min=b;
  max=r; if(g>max) max=g; if(b>max) max=b;
  V = max/255.;

  delta = max - min;
  S = delta? delta/max: 0;

  H = -1;
  if(delta) {
    if (r == max)
      H = (g - b) / delta;	/* between yellow & magenta */
    else if (g == max)
      H = 2 + (b - r) / delta;	/* between cyan & yellow */
    else
      H = 4 + (r - g) / delta;	/* between magenta & cyan */

    H *= 60;			/* degrees */
    if (H < 0)
      H += 360;
  }

  return HSXcell(H,S,V);
}

static int HSVlincell(int r, int g, int b) {
 return allHSVcell(r, g, b, HSXcelllinear);
}

static int HSVvolcell(int r, int g, int b) {
 return allHSVcell(r, g, b, HSVcellvolume);
}


/*
 * HSL space 
 */

static int HSLNcells(int NH, int NS, int NL) {
  N1=NH, N2=NS, N3=NL;
  return ((NL)*(1+((NS)-1)*(NH)));
}


static int allHSLcell(int r, int g, int b, int (*HSXcell)(double H, double S, double L)) {
  int min, max;
  double delta, H, S, L;

  min=r; if(g<min) min=g; if(b<min) min=b;
  max=r; if(g>max) max=g; if(b>max) max=b;
  L = (max+min)/(255.*2);

  delta = max - min;
  S = !delta? 0: L<0.5? delta/(max+min): delta/(2*255-max-min);

  H = -1;
  if(delta) {
    if (r == max)
      H = (g - b) / delta;	/* between yellow & magenta */
    else if (g == max)
      H = 2 + (b - r) / delta;	/* between cyan & yellow */
    else
      H = 4 + (r - g) / delta;	/* between magenta & cyan */

    H *= 60;			/* degrees */
    if (H < 0)
      H += 360;
  }
  return HSXcell(H,S,L);
}

static int HSLlincell(int r, int g, int b) {
 return allHSLcell(r, g, b, HSXcelllinear);
}

static int HSLvolcell(int r, int g, int b) {
 return allHSLcell(r, g, b, HSLcellvolume);
}


/*
 * HDS
 * 
 * (Hue,Diff,Sum) Histogram a la MPEG 7
 *
 * Daniel Heesch's explanation: The new HMMD (hue, min, max, diff) colour
 * space, which is supported by MPEG-7, is readily derived from the HSV
 * and RGB spaces. The hue component is the same as in the HSV space, and
 * max and min denote the maximum and minimum among the R, G, and B
 * values, respectively. The diff component is defined as the difference
 * between max and min. Three components are sufficient to uniquely locate
 * a point in the colour space and thus the space is effectively
 * three-dimensional. Following the MPEG-7 standard, we quantise the HMMD
 * non-uniformly into 184 bins with the three dimensions being hue, sum
 * and diff sum being defined as (max+min)/2. For details about
 * quantisation see  [B S Manjunath and J-R Ohm, 2001] . Two descriptors
 * are defined with respect to the HMMD colour space. The first is a
 * standard global histogram, the second, CSD or colour structure
 * descriptor, is described in more detail below. 
 *
 * CSD: Colour Structure Descriptor
 * 
 * This descriptor lends itself well to capturing local colour structure
 * in an image. A 8x8 structuring window is used to slide over the image.
 * Each of the 184 bins of the HMMD histogram contains the number of
 * window positions for which there is at least one pixel falling into the
 * bin under consideration. This descriptor is capable of discriminating
 * between images that have the same global colour distribution but
 * different local colour structures. Although the number of samples in
 * the 8x8 structuring window is kept constant (64), the spatial extent of
 * the window differs depending on the size of the image. Thus, for larger
 * images appropriate sub-sampling is employed to keep the total number of
 * samples per image roughly constant. The bin values are normalised by
 * dividing by the number of locations of the structuring window and fall
 * in the range [0.0, 1.0] (see [B S Manjunath and J-R Ohm, 2001] for
 * details).
 *
 * B S Manjunath and J-R Ohm, Color and Texture Descriptors,  IEEE
 * Transactions on circuits and systems for video technology, 11, pp
 * 703-715, 2001
 *
 */

#define NDB 5

static int nsb[NDB] = { 8, 4,  4,  4,  2};
static int nhb[NDB] = { 1, 8, 12, 12, 24};

static int baseDiffBin[NDB+1], Nh;

static int HDSNcells(int Nunused1, int Nunused2, int Nunused3) {
  if(!Nh) {
    int k;

    for(k=1; k<NDB+1; k++)
      baseDiffBin[k] = baseDiffBin[k-1] + nsb[k-1]*nhb[k-1];
    Nh = baseDiffBin[NDB];
  }
  return Nh;
}

static int HDScell(int r, int g, int b) {
  int min, max, nd, nh, ns, ret;
  double hue, diff, sum;

  if(!Nh) HDSNcells(0,0,0);

  min=r; if(g<min) min=g; if(b<min) min=b;
  max=r; if(g>max) max=g; if(b>max) max=b;

  diff =  max-min;
  sum  = (max+min)/2.;

  hue = 0;
  if(diff) {
    if (r == max)
      hue = (g - b) / diff;     /* between yellow & magenta */
    else if (g == max)
      hue = 2 + (b - r) / diff; /* between cyan & yellow */
    else
      hue = 4 + (r - g) / diff; /* between magenta & cyan */

    hue *= 60;			  /* degrees */
    if (hue < 0)
      hue += 360;
  }

  n2 = nd = diff/255.*NDB*.99999999;
  if(nd<0 || nd>=NDB)
    fatal("HDS conversion: nd out of range %d", nd);
  n1 = nh = hue/360.*nhb[nd]*.99999999;
  n3 = ns = sum/255.*nsb[nd]*.99999999;
  if(ns<0 || ns>=nsb[nd] || nh<0 || nh>=nhb[nd]) {
    printf("%6g %6g %6g %d %d %d\n", hue, diff, sum, nh, nd, ns);
    fatal("HDS conversion: out of range?");
  }
  ret = baseDiffBin[nd] + nh*nsb[nd] + ns;
  if(ret<0 || ret >= Nh)
    fatal("[nh=%d, nd=%d, ns=%d]\nindex %dfor HDS out of range?", nh, nd, ns, ret);
  return ret;
}
