/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * convolution.c
 *
 * Authors:
 * Marcus Pickering, February 2001: main program
 * Stefan Rueger, Aug 2002: slight changes to include this into annotate.c
 * 
 * 5 convolution filtes ala
 * K. Tieu and  P. Viola. Boosting Image Retrieval. In Proceedings of IEEE
 * Conference on Computer Vision and Pattern Recognition, 2000.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"

typedef struct {
    double scale;
    int type;
    double values[9];
} Filter;

static Filter XfilterBank[5] = {
    {4, 31,
     {1, 2, 1,}},

    {2, 31,
     {1, 0, -1,}},

    {4, 31,
     {1, -2, 1,}},

    {4, 33,
     {0, 1, 0,
      1, 0, -1,
      0, -1, 0,}},

    {4, 33,
     {0, -1, 0,
      1, 0, -1,
      0, 1, 0,}},
};

static Filter YfilterBank[5] = {
    {4, 13,
     {1, 2, 1,}},

    {2, 13,
     {1, 0, -1,}},

    {4, 13,
     {1, -2, 1,}},

    {4, 33,
     {0, 1, 0,
      -1, 0, 1,
      0, -1, 0,}},

    {4, 33,
     {0, 1, 0,
      1, 0, -1,
      0, -1, 0,}},
};

static float features[25 * 25 * 25 * 3];
static int featureIndex;

static void Convolve(double *a_pdInputBuffer, int width, int height, int a_iLevel)
{
    Filter *currX, *currY;
    double *pdYBuffer = clrmalloc(width * height * sizeof *pdYBuffer);
    double *pdOBuffer = clrmalloc(width * height * sizeof *pdOBuffer);
    double sum=0;
    int xFilter, yFilter, xOutput, yOutput, x, y;

    for (yFilter = 0; yFilter < 5; yFilter++) {
        currY = YfilterBank + yFilter;

        for (yOutput = 1; yOutput < (height - 1); yOutput++)
            for (xOutput = 1; xOutput < (width - 1); xOutput++) {
                switch (currY->type) {
                case 13:
                case 31:
                    sum = currY->values[0] * (a_pdInputBuffer[xOutput + ((yOutput - 1) * width)])
                        + currY->values[1] * (a_pdInputBuffer[xOutput + (yOutput * width)])
                        + currY->values[2] * (a_pdInputBuffer[xOutput + ((yOutput + 1) * width)]);
                    break;
                case 33:
                    sum = currY->values[1] * (a_pdInputBuffer[xOutput + ((yOutput - 1) * width)])
                        + currY->values[3] * (a_pdInputBuffer[xOutput - 1 + (yOutput * width)])
                        + currY->values[5] * (a_pdInputBuffer[xOutput + 1 + (yOutput * width)])
                        + currY->values[7] * (a_pdInputBuffer[xOutput + ((yOutput + 1) * width)]);
                    break;
                default:
                    fatal("no such type %d in Convolve()", currY->type);
                }

                /* no rectification at this step */
                pdYBuffer[(xOutput - 1) + ((yOutput - 1) * width)] = sum / currY->scale;
            }

        for (xFilter = 0; xFilter < 5; xFilter++) {
            currX = XfilterBank + xFilter;

            if (xFilter == 4 && yFilter == 4 && pars.convtype)
                break;

            for (yOutput = 1; yOutput < (height - 3); yOutput++)
                for (xOutput = 1; xOutput < (width - 3); xOutput++) {

                    switch (currX->type) {
                    case 31:
                    case 13:
                        sum = currX->values[0] * (pdYBuffer[xOutput - 1 + (yOutput * width)])
                            + currX->values[1] * (pdYBuffer[xOutput + (yOutput * width)])
                            + currX->values[2] * (pdYBuffer[xOutput + 1 + (yOutput * width)]);
                        break;
                    case 33:
                        sum = currX->values[1] * (pdYBuffer[xOutput + ((yOutput - 1) * width)])
                            + currX->values[3] * (pdYBuffer[xOutput - 1 + (yOutput * width)])
                            + currX->values[5] * (pdYBuffer[xOutput + 1 + (yOutput * width)])
                            + currX->values[7] * (pdYBuffer[xOutput + ((yOutput + 1) * width)]);
                        break;
                    default:
                        fatal("no such type %d in Convolve()", currX->type);
                    }

                    pdOBuffer[(xOutput - 1) + ((yOutput - 1) * width)]
                        = fabs(sum / currX->scale);
                }

            if (a_iLevel != pars.number[0] - 1) {
                /* The downsampling operation */
                for (y = 0; y < ((height - 4) / 2); y++)
                    for (x = 0; x < ((width - 4) / 2); x++)
                        pdOBuffer[x + (y * ((width - 4) / 2))] =
                            (pdOBuffer[(2 * x) + (2 * y * width)] +
                             pdOBuffer[(2 * x + 1) + (2 * y * width)] + pdOBuffer[(2 * x) + ((2 * y + 1) * width)] + pdOBuffer[(2 * x + 1) + ((2 * y + 1) * width)])
                            / 4.0;
                Convolve(pdOBuffer, ((width - 4) / 2), ((height - 4) / 2), a_iLevel + 1);
            } else {
                /* The downsampling operation */
                for (sum = 0, y = 0; y < ((height - 4) / 2); y++)
                    for (x = 0; x < ((width - 4) / 2); x++) {
                        sum +=
                            (pdOBuffer[(2 * x) + (2 * y * width)] +
                             pdOBuffer[(2 * x + 1) + (2 * y * width)] + pdOBuffer[(2 * x) + ((2 * y + 1) * width)] + pdOBuffer[(2 * x + 1) + ((2 * y + 1) * width)]);
                    }
                /* Normalisation */
                features[featureIndex++] = sum / ((height - 4) * (width - 4));
            }
        }
    }

    efree(pdYBuffer);
    efree(pdOBuffer);
}

static int colstodo[3];

int convolutioncheck()
{
    static char csp[4], *cspaces[2] = { "RGB", "Ybr" };
    int i, c, j, m;
    char type;

    type = pars.ap->fname[0];
    if (type == 'c') {
        pars.convtype = 0;
        printf("This is an old version of convolution that contains duplicates, it is recommended to use konvolution\n");
    } else {
        pars.convtype = 1;
    }

    if (pars.number[0] < 1 || pars.number[0] > 3)
        fatal("level of convolution should be 1, 2 or 3, but not %d", pars.number[0]);

    for (i = 0; i < sizeof cspaces / sizeof *cspaces; i++)
        for (c = 1; c < 8; c++) {
            for (m = 1, j = 0; j < 3; m <<= 1, j++)
                colstodo[j] = c & m, csp[j] = colstodo[j] ? cspaces[i][j] : '-';
            if (!strcasecmp(csp, pars.space[0])) {
                pars.useYCbCr = i;
                strcpy(pars.space[0], csp);
                return 1;
            }
        }

    fatal("either use '-a convolution-<L>-RGB' or '-a convolution-<L>-Ybr' (for YCbCr space), "
          "but not '-a convolution-%d-%s'. Each letter in RGB or Ybr can be replaced by a "
          "dash to indicate that no convolution features be computed for this component.", pars.number[0], pars.space[0]);

    return 1;
}

int convolution(FILE * fp, AImage im)
{
    int i, n, col;
    double *colextract;
    JSAMPLE *p;

    if (im.w > 0 && im.h > 0) {
        for (featureIndex = col = 0; col < 3; col++)
            if (colstodo[col]) {
                colextract = clrmalloc(im.w * im.h * sizeof *colextract);
                for (i = 0, n = im.w * im.h, p = im.cont + col; i < n; i++, p += 3)
                    colextract[i] = *p;
                Convolve(colextract, im.w, im.h, 0);
                efree(colextract);
            }
    }

    fwrite(features, sizeof *features, featureIndex, fp);

    return 1;
}
