#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fftw.h>
#include "annotate.h"

#define RED 0.3
#define GREEN 0.59
#define BLUE 0.11
#define shift(a,b) pow(-1,(a+b))

struct {
    fftw_complex **bank;
    int nx, ny;
} static FB = { NULL, 0, 0 };

enum { WINDOW = 99, SLIDE = 10, FEATURES = 2 };
static int scale, orientation, window, slide, flag = 0;
static double Ul = 0.1, Uh = 0.4;


AImage realImage(fftw_complex * data, int x, int y);
static void createFilterBank();
static int freeFilterBank();
static void applyFilter(fftw_complex * fuv, fftw_complex * huv, fftw_complex * guv, int nx, int ny);
static void calcFeatures(fftw_complex * data, int width, int height, float *mean, float *sd);

/* ------------------------------------------------------------------------------------------------------
The Gabor function generates a Gabor filter with the selected index 's' and 'n' (scale and orientation, 
respectively) from a Gabor filter bank. This filter bank is designed by giving the range of spatial 
frequency (Uh and Ul) and the total number of scales and orientations used to partition the spectrum. 

The returned filter is stored in 'Gr' (real part) and 'Gi' (imaginary part).

This fuction based on code from Manjunath
--------------------------------------------------------------------------------------------------------*/
void GaborFilter(double *Gr, double *Gi, int size, int s, int n, double Ul, double Uh, int scale, int orientation, int flag)
{
    double base, a, u0, z, Uvar, Vvar, Xvar, Yvar, X, Y, G, t1, t2, m;
    int x, y, side;

    base = Uh / Ul;
    a = pow(base, 1.0 / (double) (scale - 1));

    u0 = Uh / pow(a, (double) scale - s);

    Uvar = (a - 1.0) * u0 / ((a + 1.0) * sqrt(2.0 * log(2.0)));

    z = -2.0 * log(2.0) * (Uvar * Uvar) / u0;
    Vvar = tan(M_PI / (2 * orientation)) * (u0 + z) / sqrt(2.0 * log(2.0) - z * z / (Uvar * Uvar));

    Xvar = 1.0 / (2.0 * M_PI * Uvar);
    Yvar = 1.0 / (2.0 * M_PI * Vvar);

    t1 = cos(M_PI / orientation * (n - 1.0));
    t2 = sin(M_PI / orientation * (n - 1.0));

    side = (int) size / 2;

    for (x = 0; x < size; x++) {
        for (y = 0; y < size; y++) {
            X = (double) (x - side) * t1 + (double) (y - side) * t2;
            Y = (double) -(x - side) * t2 + (double) (y - side) * t1;
            G = 1.0 / (2.0 * M_PI * Xvar * Yvar) * pow(a, (double) scale - s) * exp(-0.5 * ((X * X) / (Xvar * Xvar) + (Y * Y) / (Yvar * Yvar)));
            Gr[x * size + y] = G * cos(2.0 * M_PI * u0 * X);
            Gi[x * size + y] = G * sin(2.0 * M_PI * u0 * X);
        }
    }
    if (flag == 1) {
        m = 0;
        for (x = 0; x < size; x++)
            for (y = 0; y < size; y++)
                m += Gr[x * size + y];

        m /= pow((double) size, 2.0);

        for (x = 0; x < size; x++)
            for (y = 0; y < size; y++)
                Gr[x * size + y] -= m;
    }
}

int gabor(FILE * fp, AImage im)
{
    int nx, ny, i, j, f, index;
    unsigned char *c;
    float *result, mean, sd;
    //AImage restored;
    fftw_complex *fuv, *img;
    fftw_complex *huv, *guv, *out;
    fftwnd_plan pImage, pout;

    // set up the feature vector
    result = clrmalloc(scale * orientation * FEATURES * sizeof *result);

    // Deal with the Image
    // calc nx and ny, read in the image, and fft it to fuv = f(u,v) 
    nx = im.h - (im.h + 1) % 2;
    ny = im.w - (im.w + 1) % 2;

    fuv = fftw_malloc(sizeof *fuv * nx * ny);
    img = fftw_malloc(sizeof *img * nx * ny);

    c = im.cont;
    for (i = 0; i < im.h; i++) {
        for (j = 0; j < im.w; j++, c += 3) {
            if (i < nx && j < ny) {
                img[i * ny + j].re = (double) ((c[0] * RED + c[1] * GREEN + c[2] * BLUE));      //* shift(i,j));
                img[i * ny + j].im = 0.0;
            }
        }
    }
    pImage = fftw2d_create_plan(nx, ny, FFTW_FORWARD, FFTW_ESTIMATE);
    fftwnd_one(pImage, img, fuv);
    fftw_free(img);
    fftwnd_destroy_plan(pImage);
    // Image DONE...

    // Create the filter Bank - first check to see if the existing one is ok...
    if (FB.nx != nx || FB.ny != ny) {
        if (FB.bank)
            freeFilterBank();
        createFilterBank(ny, nx);
    }
    // filter the image with each filter in the bank...
    // first create the plan

    guv = fftw_malloc(sizeof *guv * nx * ny);
    out = fftw_malloc(sizeof *out * nx * ny);
    pout = fftw2d_create_plan(nx, ny, FFTW_BACKWARD, FFTW_ESTIMATE);
    for (f = 0; f < scale * orientation; f++) {
        huv = *(FB.bank + f);
        // multiply the transformed image and the filter to give another fftw_complex
        // and then inverse fft to give the result...
        for (i = 0; i < nx; i++) {
            for (j = 0; j < ny; j++) {
                index = i * ny + j;
                guv[index].re = (fuv[i * ny + j].re * huv[index].re - fuv[i * ny + j].im * huv[index].im) * shift(i, j);
                guv[index].im = (fuv[i * ny + j].re * huv[index].im + fuv[i * ny + j].im * huv[index].re) * shift(i, j);
            }
        }
        fftwnd_one(pout, guv, out);
        calcFeatures(out, ny, nx, &mean, &sd);
        result[2 * f] = mean;
        result[2 * f + 1] = sd;
        //printf("filter=%d mean=%f sd=%f\n", f, result[2*f], result[2*f +1]); 
        //restored = realImage(out, nx, ny);
        //sprintf(name, "gabor_%d.Jpg", f);
        //fpg = fopen(name,"w");
        //WriteImageFile(restored, fpg);
        //fclose(fpg);
        //printf("Written file\n");


    }

    fwrite(result, sizeof *result, scale * orientation * FEATURES, fp);
    free(result);
    //fftw_execute(p2);

    fftwnd_destroy_plan(pout);
    //fftw_destroy_plan(pout);
    fftw_free(out);
    fftw_free(guv);
    fftw_free(fuv);
    return 1;
}

/*
 * This function uses a bank of fixed size filters and slides them over the
 * image, features are calculted for each filtered image and averaged...
 */
int histGabor(FILE * fp, AImage im)
{
    int startX, startY, x, y, i, j, f, total = 0;
    float *result, mean, sd;
    AImage gIm;
    fftw_complex *fuv, *img, *guv, *out;
    fftwnd_plan pImage, pOut;
    if (window > im.h || window > im.w)
        fatal("The window size of %d is larger than the image %d x %d", window, im.h, im.w);

    // create filterBank
    if (!FB.bank)
        createFilterBank(window, window);

    result = clrmalloc(scale * orientation * FEATURES * sizeof *result);
    gIm = ConvertToGrey(im, 256);
    fuv = fftw_malloc(sizeof *fuv * window * window);
    img = fftw_malloc(sizeof *img * window * window);
    guv = fftw_malloc(sizeof *guv * window * window);
    out = fftw_malloc(sizeof *out * window * window);
    pImage = fftw2d_create_plan(window, window, FFTW_FORWARD, FFTW_ESTIMATE);
    pOut = fftw2d_create_plan(window, window, FFTW_BACKWARD, FFTW_ESTIMATE);

    //printf("h=%d\tw=%d\n", im.h, im.w);
    for (startX = 0; startX <= im.h - window; startX += slide) {
        for (startY = 0; startY <= im.w - window; startY += slide) {
            // populate img
            for (x = startX, i = 0; x < startX + window; x++, i++) {
                for (y = startY, j = 0; y < startY + window; y++, j++) {
                    img[i * window + j].re = gIm.cont[x * window + y];
                    img[i * window + j].im = 0;
                }
            }
            //      printf("sX=%d\tsY=%d\n", startX, startY);
            fftwnd_one(pImage, img, fuv);
            // filter F(u,v) with each of the filters in the bank
            for (f = 0; f < scale * orientation; f++) {
                applyFilter(fuv, FB.bank[f], guv, window, window);
                fftwnd_one(pOut, guv, out);
                // calculate the feature
                calcFeatures(out, window, window, &mean, &sd);
                result[FEATURES * f] += mean;
                result[FEATURES * f + 1] += sd;
            }
            total++;
        }
    }
    for (f = 0; f < scale * orientation * FEATURES; f++) {
        result[f] /= total;
        //printf("result%d\t= %f\n", f, result[f]);
    }

    fwrite(result, sizeof *result, scale * orientation * FEATURES, fp);
    efree(result);
    // do not free the filterBank - it will be used each time the function is called
    // it should be cleaned up by recusive.c

    fftwnd_destroy_plan(pImage);
    fftwnd_destroy_plan(pOut);
    efree(gIm.cont);
    fftw_free(fuv);
    fftw_free(img);
    fftw_free(guv);
    fftw_free(out);
    return 1;
}


int gaborCheck(void)
{
    scale = pars.number[0];
    if (scale < 2 || scale > 6)
        fatal("Scale should be between 2 and 6 - default value is 4\n");

    orientation = pars.number[1];
    if (orientation < 1 || orientation > 6)
        fatal("Orientation should be between 1 and 6 - default value is 4\n");

    //printf("scale=%d orientation=%d\n", scale, orientation);
    return 1;
}

int winGaborCheck(void)
{
    window = pars.number[2];
    if (window < 5 || window > 255)
        fatal("Window should be between 5 and 255 - default value is 65\n");
    slide = pars.number[3];
    if (slide < 1)
        fatal("Slide value should be greater than 0 - default value is 10\n");

    //printf("window=%d slide=%d\n", window, slide);

    return gaborCheck();
}


static void applyFilter(fftw_complex * fuv, fftw_complex * huv, fftw_complex * guv, int nx, int ny)
{
    int i, j, index;
    double shift;
    for (i = 0; i < nx; i++) {
        for (j = 0; j < ny; j++) {
            index = i * ny + j;
            shift = shift(i, j);
            guv[index].re = (fuv[index].re * huv[index].re - fuv[index].im * huv[index].im) * shift;
            guv[index].im = (fuv[index].re * huv[index].im + fuv[index].im * huv[index].re) * shift;
        }
    }
}

/*
 * Function to save the real part of an fftw_complex file as a jpg file
 * used to check what on earth is going on!
 */
AImage realImage(fftw_complex * data, int x, int y)
{
    int i, j, n;
    double real, imag, abs, max, factor;
    AImage ret;
    n = 0;
    max = 0;
    ret.w = y;
    ret.h = x;

    for (i = 0; i < ret.h; i++)
        for (j = 0; j < ret.w; j++) {
            real = data[i * y + j].re;
            imag = data[i * y + j].im;
            abs = sqrt(real * real + imag * imag);
            if (max < abs)
                max = abs;
        }
    factor = 255 / max;

    ret.cont = clrmalloc(x * y * 3 * sizeof *ret.cont);
    for (i = 0; i < x; i++) {
        for (j = 0; j < y; j++) {
            real = data[i * y + j].re;
            imag = data[i * y + j].im;
            abs = sqrt(real * real + imag * imag);
            ret.cont[n + 1] = (unsigned int) (abs * factor);
            //printf("%d ", ret.cont[n+1]);
            n += 3;
        }
    }
    //printf("max=%f min=\n", max);
    return ret;
}

static void createFilterBank(int width, int height)
{
    int size, s, n, i, j, borderW, borderH, index;
    fftw_complex *filter, *huv, *huvInitial;
    fftwnd_plan p;
    double *Gr, *Gi;
    FB.nx = height;
    FB.ny = width;
    //printf("creating filter bank... w=%d h=%d\n", width, height);

    size = width < height ? width : height;
    FB.bank = clrmalloc(scale * orientation * sizeof *FB.bank);
    // 888888888888888
    Gr = clrmalloc(size * size * sizeof *Gr);
    Gi = clrmalloc(size * size * sizeof *Gr);
    //CreateMatrix(&Gr, size, size);
    //CreateMatrix(&Gi, size, size);
    // 888888888888888
    huvInitial = fftw_malloc(height * width * sizeof(fftw_complex));
    filter = fftw_malloc(height * width * sizeof *filter);
    p = fftw2d_create_plan(height, width, FFTW_FORWARD, FFTW_ESTIMATE);

    for (s = 0; s < scale; s++) {
        for (n = 0; n < orientation; n++) {
            // 888888888888888
            GaborFilter(Gr, Gi, size, s + 1, n + 1, Ul, Uh, scale, orientation, flag);
            // 888888888888888
            borderW = (width - size) / 2;
            borderH = (height - size) / 2;
            //printf("width=%d height=%d bW=%d bH=%d size=%d side=%d\n", width, height, borderW, borderH, size, (size-1)/2);      
            for (i = 0; i < height; i++) {
                for (j = 0; j < width; j++) {
                    if ((j < borderW) || (j > (borderW + size - 1)) || (i < borderH) || (i > (borderH + size - 1))) {
                        filter[width * i + j].re = 0;
                        filter[width * i + j].im = 0;
                    } else {
                        index = (i - borderH) * size + j - borderW;
                        filter[width * i + j].re = Gr[index];   // * shift(i-borderH,j-borderW);
                        filter[width * i + j].im = Gi[index];   // * shift(i-borderH,j-borderW);
                    }
                }
            }
            huv = fftw_malloc(height * width * sizeof(fftw_complex));
            fftwnd_one(p, filter, huv);
            *(FB.bank + s * orientation + n) = huv;
            //printf("huv=%p\n", huv);
        }
    }
    // 888888888888888 Destroy the matricies...
    efree(Gr);
    efree(Gi);
    fftwnd_destroy_plan(p);
    fftw_free(filter);
    fftw_free(huvInitial);
    //printf("finished...\n");
}

static int freeFilterBank()
{
    int i, size;
    size = scale * orientation;
    //printf("size=%d\n", size);
    for (i = 0; i < size; i++) {
        free(*(FB.bank + i));
    }
    free(FB.bank);
    return 1;
}

static void calcFeatures(fftw_complex * data, int width, int height, float *mean, float *sd)
{
    int i, j;
    double abs, sum, total, sum_2, area, var;
    sum = 0;
    total = 0;
    sum_2 = 0;
    area = width * height;
    for (i = 0; i < height; i++) {
        for (j = 0; j < width; j++) {
            abs = (double) (sqrt(pow(data[i * width + j].re, 2) + pow(data[i * width + j].im, 2)));
            sum += abs;
            sum_2 += pow(abs, 2);
            total++;
        }
    }
    *mean = (float) sum / (total * area);
    var = (total * sum_2 - sum * sum) / (total * total);
    var = var > 0 ? var : 0.0;
    *sd = (float) sqrt(var) / area;
    //printf("tot=%f, sum=%f, sum_2=%f, area=%f\t", total, sum, sum_2, area); 
    //printf("%f\n", *sd);
}
