/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * margcolourhist.c
 *
 * Creates various marginal colour histograms for an image or, alternatively,
 * the central moments of the full 256-bin histogram of each colour chanel.
 * More, precicely, we compute the nth root of the first n central moments.
 *
 * Stefan M Rueger
 *
 * For each marginal colour model we support several major methods:
 *
 * G for global
 * computes a global normalised colour histogram for each colour component
 * The sum of values for each component is 1/3 so that the overall sum of
 * feature components is 1.
 *
 *
 * F (for focus)
 * computes a global normalised colour histogram for each colour component
 * in the 25% centre area of the image.
 *
 *
 * S (for structured)
 * computes a structured colour histogram for each component
 *
 * For each pixel-channel in the image the correspeonding bin is computed.
 * A gliding 8x8 window  moves over the image - each histogram bin counts
 * the number of windows which contain at least one pixel of the
 * corresponding bin. Local information is folded into this type of
 * histogram. Normalisation is through dividing by the number of different
 * window positions; each component is thus between 0 and 1, but the sum
 * can be as large as 64 (in rare cases).
 *
 *
 * C (for centre)  .-------------------.
 *                 |  0      |      2  |
 *                 |   .-----------.   |
 *                 |   |           |   |
 *                 |---|     4     |---|
 *                 |   |           |   |
 *                 |   .-----------.   |
 *                 |  1      |      3  |
 *                 .-------------------.
 * splits the image into 5 areas: centre (50% area) and border, which is 
 * divided into 4 quarters.
 *
 * For each area a percentage histogram for is computed: For example in
 * RGB there are NR+NG+NB 1-d colour bins and consequently 5*(NR+NG+NB)
 * float numbers which make up the feature vector. The central histogram
 * is weighted with a relative factor of 16, so that (considering each of
 * the 4 border tiles weighs 1) its contribution is 80% of the image. The
 * four border tiles have a cumulative importance  of 20%. The sum of all
 * the components is made to be 1.
 * 
 *
 * L (for local)   .-------------------.
 *                 |                   |
 *                 |   .-----------.   |
 *                 |   |  0  |  2  |   |
 *                 |   |-----+-----|   |
 *                 |   |  1  |  3  |   |
 *                 |   .-----------.   |
 *                 |                 4 |
 *                 .-------------------.
 * splits the image into 5 areas: border (50% area) and the centre which is
 * divided into 4 quarters.
 *
 * For each area a percentage histogram for each colour channel is
 * computed:
 * For example in RGB there are NR+NG+NB 1-d colour bins and consequently
 * 5*(NR+NG+NB) float numbers which make up the feature vector. As each
 * area histogram is normalised separately, the middle histograms are
 * weighed higher than the border; they contribute 20% each to the feature
 * vector (but they have only 12.5% area) The border is demoted from 50%
 * area to only 20% contribution. The sum of all the components is made to
 * be 1.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>

#include "annotate.h"
#include "margcolourmodels.c"

static struct {
  void (*realsetn123)(colour_t, colour_t, colour_t); void (*setn123)(colour_t, colour_t, colour_t);
  double cmin[3], cmax[3];
  Coord *(*smoothing)(int, int, int);
} args;

static Pixel *u3_m;
#define CACHESIZ (256*256*256)

void u3_cache(colour_t r, colour_t g, colour_t b) {
  int ind = (((r<<8)+g)<<8)+b;
  if(!u3_m[ind].n[0])
    args.realsetn123(r,g,b), u3_m[ind] = pix;
  else
    pix = u3_m[ind];
}

static int 
  margglobalcolourhist(FILE *fp, AImage im),
  margfocuscolourhist(FILE *fp, AImage im),
  margstructuredcolourhist(FILE *fp, AImage im),
  margcentrecolourhist(FILE *fp, AImage im),
  marglocalcolourhist(FILE *fp, AImage im),
  (*torun)(FILE *fp, AImage im);

static int Hx[3], Ncol, Mx[3], mom;

static int runmarghistogram(FILE *fp, AImage im, void (*setn123)(colour_t, colour_t, colour_t)) {
  args.setn123 = setn123;
  args.smoothing = pars.gridS? smoothing_S: pars.gridG? smoothing_G: NULL;
  if(!args.realsetn123) {
    int size;
    if(!pars.flag.C) {
      if((size=CACHESIZ*sizeof*u3_m) < pars.memleft && (u3_m = malloc(size))) {
        memset(u3_m,0,size), args.realsetn123 = args.setn123, args.setn123 = u3_cache, pars.memleft -= size; // printf("uc cache\n");
        if(pars.flag.d) printf("cache 3\n");
      }
    }
  }
  return torun(fp, im);
}

void writehist(float *hist, int j, FILE *fp) {
  int m = Mx[j], n = Hx[j];

  if(!mom) {                    /* colour histograms, not colour moments */
    fwrite(hist, sizeof *hist, n, fp);
    return;
  }

  if(m>0) {
    double cmin=args.cmin[j], cmax=args.cmax[j], scale;
    double moments[20], mass, x, xi;
    float  moms[sizeof moments/sizeof*moments];
    int i, j;

    if(m>sizeof moments/sizeof*moments)
      fatal("cannot do colour moments of order %d; recompile with changed constant in writehist()");

    for(j=0; j<m; j++)          /* initialise */
      moments[j] = 0;

    for(mass=i=0; i<n; i++)     /* average first */
       moments[0] += hist[i]*i, mass += hist[i];
    if(mass)
      moments[0] /= mass;

    for(i=0; i<n; i++)          /* central moments */
      for(x=xi=(i-moments[0]), j=1; j<m; j++)
        x *= xi, moments[j] += hist[i]*x;
    if(mass)
      for(j=1; j<m; j++)
        moments[j] /= mass;

    for(moms[0]=moments[0], j=1; j<m; j++) /* take corresponding root */
      moms[j] = pow(fabs(moments[j]), 1./(j+1))*(moments[j]<0? -1: 1);

    if((scale=(cmax-cmin)/256) != 0) {     /* shift to range of colourmodel */
      for(j=0; j<m; j++) 
        moms[j] = scale*moms[j];
      moms[0] += cmin;
    }

    fwrite(moms, sizeof*moms, m, fp);
  }
}

#define NT 5 /* tiles */
#define BF 0.14644660940673 /* ((1-sqrt(.5))/2) such that border = 50%  */

static int marglocalcolourhist(FILE *fp, AImage im) {
  int i, j, k, ci, r, g, b;
  float c_hist[3][NT][256], d_hist[256], n_hist;
  unsigned char *c;
  Coord *co;
  int xl=BF*im.w+1, xr=(1-BF)*im.w, yt=BF*im.h+1, yb=(1-BF)*im.h, xm=im.w/2, ym=im.h/2;

  memset(c_hist, 0, sizeof c_hist);

  if(args.smoothing) { // smooth each colour pixel with a gaussian function
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        k = i<xl || i>xr || j<yt || j>yb? /* border */ 4: /* 4 inner */ (i>xm)*2 + (j>ym);
        co = args.smoothing(c[0], c[1], c[2]);
        for(r=co->r, g=co->g, b=co->b, ci=0; ci<co->N; ci++) {
          args.setn123(r+co->c[ci].dr, g+co->c[ci].dg, b+co->c[ci].db); // sets n1, n2, n3
          c_hist[0][k][n1] += co->c[ci].w;
          c_hist[1][k][n2] += co->c[ci].w;
          c_hist[2][k][n3] += co->c[ci].w;
        }
      }
  } else
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        k = i<xl || i>xr || j<yt || j>yb? /* border */ 4: /* 4 inner */ (i>xm)*2 + (j>ym);
        args.setn123(c[0], c[1], c[2]); // sets n1, n2, n3;
        c_hist[0][k][n1]++;
        c_hist[1][k][n2]++;
        c_hist[2][k][n3]++;
      }

  for(k=0; k<NT; k++) 
    for(j=0; j<3; j++)
      if(Hx[j] > 0) {
        memset(d_hist, 0, Hx[j]*sizeof*d_hist);
        for(n_hist=0, i=0; i<256; i++)
          n_hist+=c_hist[j][k][i];
        if(n_hist)
          for(i=0; i<256; i++)
            d_hist[i*Hx[j]/256] += c_hist[j][k][i]/(Ncol*NT*n_hist); // uniform weighing
         writehist(d_hist, j, fp);
      }

  return 1;
}


static int margcentrecolourhist(FILE *fp, AImage im) {
  int i, j, k, ci, r, g, b;
  float c_hist[3][NT][256], d_hist[256], n_hist;
  unsigned char *c;
  Coord *co;
  int xl=BF*im.w+1, xr=(1-BF)*im.w, yt=BF*im.h+1, yb=(1-BF)*im.h, xm=im.w/2, ym=im.h/2;

  memset(c_hist, 0, sizeof c_hist);

  if(args.smoothing) { // smooth each colour pixel with a gaussian function
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        k = i>=xl && i<=xr && j>=yt && j<=yb? /* centre */ 4: /* 4 outer */ (i>xm)*2 + (j>ym);
        co = args.smoothing(c[0], c[1], c[2]);
        for(r=co->r, g=co->g, b=co->b, ci=0; ci<co->N; ci++) {
          args.setn123(r+co->c[ci].dr, g+co->c[ci].dg, b+co->c[ci].db); // sets n1, n2, n3
          c_hist[0][k][n1] += k==4? 16*co->c[ci].w: co->c[ci].w;
          c_hist[1][k][n2] += k==4? 16*co->c[ci].w: co->c[ci].w;
          c_hist[2][k][n3] += k==4? 16*co->c[ci].w: co->c[ci].w; /* weigh centre more */
        }
      }
  } else
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        k = i>=xl && i<=xr && j>=yt && j<=yb? /* centre */ 4: /* 4 outer */ (i>xm)*2 + (j>ym);
        args.setn123(c[0], c[1], c[2]); // sets n1, n2, n3;
        c_hist[0][k][n1] += k==4? 16: 1;
        c_hist[1][k][n2] += k==4? 16: 1;
        c_hist[2][k][n3] += k==4? 16: 1; /* weigh centre more */
      }

  for(k=0; k<NT; k++)
    for(j=0; j<3; j++)
      if(Hx[j] > 0) {
        memset(d_hist, 0, Hx[j]*sizeof*d_hist);
        for(n_hist=0, i=0; i<256; i++)
          n_hist+=c_hist[j][k][i];
        if(n_hist /= k==4? 16: 1)
          for(i=0; i<256; i++)
            d_hist[i*Hx[j]/256] += c_hist[j][k][i]/(Ncol*20*n_hist);
        writehist(d_hist, j, fp);
      }

  return 1;
}


static int margglobalcolourhist(FILE *fp, AImage im) {
  int i, j, ci, r, g, b;
  float c_hist[3][256], d_hist[256], n_hist;
  unsigned char *c;
  Coord *co;

  memset(c_hist, 0, sizeof c_hist);

  if(args.smoothing) { // smooth each colour pixel with a gaussian function
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        co = args.smoothing(c[0], c[1], c[2]);
        for(r=co->r, g=co->g, b=co->b, ci=0; ci<co->N; ci++) {
          args.setn123(r+co->c[ci].dr, g+co->c[ci].dg, b+co->c[ci].db); // sets n1, n2, n3
          c_hist[0][n1] += co->c[ci].w;
          c_hist[1][n2] += co->c[ci].w;
          c_hist[2][n3] += co->c[ci].w;
        }
      }
  } else
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        args.setn123(c[0], c[1], c[2]); // sets n1, n2, n3;
        c_hist[0][n1]++;
        c_hist[1][n2]++;
        c_hist[2][n3]++;
      }

  for(j=0; j<3; j++)
    if(Hx[j] > 0) {
      memset(d_hist, 0, Hx[j]*sizeof*d_hist);
      for(n_hist=0, i=0; i<256; i++)
        n_hist+=c_hist[j][i];
      if(n_hist)
        for(i=0; i<256; i++)
          d_hist[i*Hx[j]/256] += c_hist[j][i]/(Ncol*n_hist);
      writehist(d_hist, j, fp);
    }

  return 1;
}

static int margfocuscolourhist(FILE *fp, AImage im) {
  int i, j, xl, xr, yt, yb, ci, r, g, b;
  float c_hist[3][256], d_hist[256], n_hist;
  unsigned char *c;
  Coord *co;

  memset(c_hist, 0, sizeof c_hist);

  xl=.25*im.w, xr=0.75*im.w, yt=0.25*im.h, yb=0.75*im.h;
  if((xr-xl)>0 && (yb-yt)>0) {
    if(args.smoothing) { // smooth each colour pixel with a gaussian function
      for(c=im.cont, j=yt; j<yb; j++)
        for(i=xl, c=im.cont+(j*im.w+i)*3; i<xr; i++, c+=3) {
          co = args.smoothing(c[0], c[1], c[2]);
          for(r=co->r, g=co->g, b=co->b, ci=0; ci<co->N; ci++) {
            args.setn123(r+co->c[ci].dr, g+co->c[ci].dg, b+co->c[ci].db); // sets n1, n2, n3
            c_hist[0][n1] += co->c[ci].w;
            c_hist[1][n2] += co->c[ci].w;
            c_hist[2][n3] += co->c[ci].w;
          }
        }
    } else
      for(c=im.cont, j=yt; j<yb; j++)
        for(i=xl, c=im.cont+(j*im.w+i)*3; i<xr; i++, c+=3) {
          args.setn123(c[0], c[1], c[2]); // sets n1, n2, n3;
          c_hist[0][n1]++;
          c_hist[1][n2]++;
          c_hist[2][n3]++;
        }
  }

  for(j=0; j<3; j++)
    if(Hx[j] > 0) {
      memset(d_hist, 0, Hx[j]*sizeof*d_hist);
      for(n_hist=0, i=0; i<256; i++)
        n_hist+=c_hist[j][i];
      if(n_hist)
        for(i=0; i<256; i++)
          d_hist[i*Hx[j]/256] += c_hist[j][i]/(Ncol*n_hist);
      writehist(d_hist, j, fp);
    }

  return 1;
}


static int margstructuredcolourhist(FILE *fp, AImage im) {
  int i, j, ci, cj, k, bin;
  float c_hist[256], n_hist;
  unsigned char *c, c_mark[256], *c_bin[3];

  if(pars.gridS || pars.gridG)
    fatal("cannot use -S structured marginal colour histogram with :S<gridsz>-... or :G<gridsz>-... smoothing modifier");

  for(k=0; k<3; k++)
    c_bin[k]  = clrmalloc(im.h*im.w*sizeof*c_bin[k]);
  for(c=im.cont, k=j=0; j<im.h; j++)
    for(i=0; i<im.w; i++, c+=3) {
      args.setn123(c[0], c[1], c[2]); // sets n1, n2, n3;
      c_bin[0][k] = (n1*Hx[0])/256;
      c_bin[1][k] = (n2*Hx[1])/256;
      c_bin[2][k] = (n3*Hx[2])/256;
      k++;
    }
  
  for(k=0; k<3; k++) {
    memset(c_hist, 0, sizeof c_hist);
    if(Hx[k]>0) {
      for(j=0; j<im.h-7; j++)
        for(i=0; i<im.w-7; i++)
           for(memset(c_mark, 0, sizeof c_mark), cj=0; cj<8; cj++)
             for(ci=0; ci<8; ci++)
               if(!c_mark[bin = c_bin[k][(j+cj)*im.w + i+ci]])
                 c_mark[bin] = 1, c_hist[bin]++;
  
      if((n_hist=(im.h-7)*(im.w-7)) > 0 && (im.h>7))
        for(i=0; i<Hx[k]; i++)
          c_hist[i]/=n_hist;

      writehist(c_hist, k, fp);
    }
    efree(c_bin[k]);
  }

  return 1;
}

int margXYZcheck() {
  int i;

  switch(*pars.space[0]) {
  case 'M': case 'm': *pars.space[0] = 'M'; mom=1; break;
  case 'H': case 'h': *pars.space[0] = 'H'; mom=0; break;
  default:
    fatal("must specify H or M in  marg<colourmodel>-N1xN2xN3-[HM]-[GCLS]; "
          "M stands for moment computation, H for histogram computation.");
  }

  for(Ncol=i=0; i<3; i++) {
    Ncol += pars.number[i]>0;
    if(mom) {
      if(pars.number[i]<0 || pars.number[i]> 10)
        fatal("moment degrees (N1, N2, N3) in marg<colourmodel>-N1+N2+N3-M-[GCLS] must be in range of [0,10]");
      Hx[i] = 256, Mx[i] = pars.number[i];
    } else {
      if(pars.number[i]<0 || pars.number[i]>256)
        fatal("number of bins (N1, N2, N3) in marg<colourmodel>-N1+N2+N3-H-[GCLS] must be in range of [0,256]");
      Hx[i] = pars.number[i];
    }
  }
  if(Ncol<=0)
    fatal("at least one number (N1,N2,N3) in marg<colourmodel>-N1+N2+N3-[HM]-[GCLS] must be positive.");

  switch(*pars.space[1]) {
  case 'G': case 'g': *pars.space[1]='G'; torun = margglobalcolourhist;     break;
  case 'F': case 'f': *pars.space[1]='F'; torun = margfocuscolourhist;      break;
  case 'S': case 's': *pars.space[1]='S'; torun = margstructuredcolourhist; break;
  case 'C': case 'c': *pars.space[1]='C'; torun = margcentrecolourhist;     break;
  case 'L': case 'l': *pars.space[1]='L'; torun = marglocalcolourhist;      break;
  default:
    fatal("must specify G, F, S, C or L in marg<colourmodel>-N1+N2+N3-[HM]-[GFSCL]. G, F, S, C, or L stand for global, focus, structured, centred or local histogram types");
  }
 return 1;
}

int margrgb(FILE *fp, AImage im) {
  args.cmin[0] = args.cmin[1] = args.cmin[2] = 0;
  args.cmax[0] = args.cmax[1] = args.cmax[2] = 1;
  pars.flag.C = 1; // no caching for r,g,b
  return runmarghistogram(fp, im, toRGB); 
}

int marghsvlin(FILE *fp, AImage im) {
  args.cmin[0] = args.cmin[1] = args.cmin[2] = 0;
  args.cmax[0] = args.cmax[1] = args.cmax[2] = 1;
  return runmarghistogram(fp, im, toHSVlin); 
}

int marghsvvol(FILE *fp, AImage im) {
  args.cmin[0] = args.cmin[1] = args.cmin[2] = 0;
  args.cmax[0] = args.cmax[1] = args.cmax[2] = 1;
  return runmarghistogram(fp, im, toHSVvol); 
}

int marghsllin(FILE *fp, AImage im) {
  args.cmin[0] = args.cmin[1] = args.cmin[2] = 0;
  args.cmax[0] = args.cmax[1] = args.cmax[2] = 1;
  return runmarghistogram(fp, im, toHSLlin); 
}

int marghslvol(FILE *fp, AImage im) {
  args.cmin[0] = args.cmin[1] = args.cmin[2] = 0;
  args.cmax[0] = args.cmax[1] = args.cmax[2] = 1;
  return runmarghistogram(fp, im, toHSLvol); 
}

int margcieluv(FILE *fp, AImage im) {
  args.cmin[0] = 0;  args.cmin[1] = -1.23324;  args.cmin[2] = -1.39267;
  args.cmax[0] = 1;  args.cmax[1] =  1.47573;  args.cmax[2] =  0.121182;
  return runmarghistogram(fp, im, toCIELUV); 
}

int margcielab(FILE *fp, AImage im) {
  args.cmin[0] = 0; args.cmin[1] = -0.861813; args.cmin[2] = -1.078620;
  args.cmax[0] = 1; args.cmax[1] = 0.982352; args.cmax[2] = 0.944758;
  return runmarghistogram(fp, im, toCIELAB); 
}

int margypcbcr(FILE *fp, AImage im) {
  args.cmin[0] = 0; args.cmin[1] = args.cmin[2] = -.5;
  args.cmax[0] = 1; args.cmax[1] = args.cmax[2] = +.5;
  return runmarghistogram(fp, im, toYpCbCr); 
}
