/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * marginal colour models (ie, 1d histograms for each colour component)
 *
 * Stefan M Rueger
 *
 * In the marginalised colour models each of the (usually) three colour
 * components is treated separately, in isolation. As a first step we can 
 * always create the summary histogram, assuming 256 bins, for each colour
 * channel separately. After this, we have 3 one-dimensional histograms,
 * which can be made coarser by 
 *   - subdividing the histograms into N1, N2, and, respectively, N3 bins.
 *   - by computing the first N1, N2, N3 statistical moments of the
 *     respective one-dimenstional histogram.
 * 
 * We assume that each colourmodel has 3 coordinates which are separately
 * to be split up into N1, N2 and N3 bins, respectively.  Treating each
 * (marginalised) colour component separately,  we usually get N1+N2+N3
 * bins, ie, we create a histogram for  each colour component on its own.
 *
 */

typedef unsigned char colour_t;

typedef struct {
    colour_t n[4];
} Pixel;
static Pixel pix = { {1, 0, 0, 0} };

#define n1 pix.n[1]
#define n2 pix.n[2]
#define n3 pix.n[3]

static void toRGB(colour_t r, colour_t g, colour_t b)
{
    n1 = r;
    n2 = g;
    n3 = b;
}


/*
 * Y'CbCr is a linear transformation from gamma-corrected (R',G',B')
 * values. The coefficients used here are for Recommendation ITU-R
 * BT.601-4,  Recommendation ITU-R BT.470-4 System B and G, SMPTE
 * 170M-1994 and SMPTE 293M-1996. 
 */

static void toYpCbCr(colour_t r, colour_t g, colour_t b)
{
    double Yp, Cb, Cr, fr = 0.299, fg = 0.587, fb = 0.114;
    int yp, cb, cr;

    Yp = fr * r / 255. + fg * g / 255. + fb * b / 255.;
    Cb = -fr / 2 / (1 - fb) * r / 255. - fg / 2 / (1 - fb) * g / 255. + 0.5 * b / 255.;
    Cr = 0.5 * r / 255. - fg / 2 / (1 - fr) * g / 255. - fb / 2 / (1 - fr) * b / 255.;

    yp = Yp * 255.999999;
    cb = (0.5 + Cb) * 255.999999;
    cr = (0.5 + Cr) * 255.999999;

    if (yp < 0 || cb < 0 || cr < 0 || yp >= 256 || cb >= 256 || cr >= 256) {
        printf("%6g %6g %6g %d %d %d\n", Yp, Cb, Cr, yp, cb, cr);
        fatal("Y'CbCr conversion: vales out of range?");
    }
    n1 = yp;
    n2 = cb;
    n3 = cr;
}


/*
 * CIELUV and CIELAB space, more perceptually uniform
 *
 *  see also http://www.cs.rit.edu/~ncs/color/ and click on t_convert.html
 * 
 *  [ X ]   [  0.412453  0.357580  0.180423 ]   [ R ]
 *  [ Y ] = [  0.212671  0.715160  0.072169 ] * [ G ]
 *  [ Z ]   [  0.019334  0.119193  0.950227 ]   [ B ]
 * 
 * 
 * 
 * L* = 116 * (Y/Yn)^(1/3) - 16    for Y/Yn > 0.008856
 * L* = 903.3 * Y/Yn               otherwise
 * 
 * a* = 500 * ( f(X/Xn) - f(Y/Yn) )
 * b* = 200 * ( f(Y/Yn) - f(Z/Zn) )
 *     where f(t) = t^(1/3)      for t > 0.008856
 *               f(t) = 7.787 * t + 16/116    otherwise
 * 
 * 
 * ---
 * 
 * L* =  116 * (Y/Yn)1/3 - 16
 * u* =  13L* * ( u' - un' )
 * v* =  13L* * ( v' - vn' )
 * 
 * The quantities un' and vn'  refer to the reference white or the light
 * source; for the 20 observer and illuminant C,  un' = 0.2009, vn' = 0.4610
 * [1]. Equations for u' and v' are given below:
 * 
 *         u' = 4X / (X + 15Y + 3Z) = 4x / ( -2x + 12y + 3 )
 *         v' = 9Y / (X + 15Y + 3Z) = 9y / ( -2x + 12y + 3 ).
 * 
 */


static void toCIELAB(colour_t r, colour_t g, colour_t b)
{
    double XbXn, YbYn, ZbZn, Ls, as, bs;
    int nl, na, nb;

#define WR 255
#define WG 255
#define WB 255

#define Xn (0.412453*WR + 0.357580*WG + 0.180423*WB)
#define Yn (0.212671*WR + 0.715160*WG + 0.072169*WB)
#define Zn (0.019334*WR + 0.119193*WG + 0.950227*WB)

    XbXn = (0.412453 * r + 0.357580 * g + 0.180423 * b) / Xn;
    YbYn = (0.212671 * r + 0.715160 * g + 0.072169 * b) / Yn;
    ZbZn = (0.019334 * r + 0.119193 * g + 0.950227 * b) / Zn;

#define fscale(x) ((x)>0.008856? pow((x),1/3.): 7.787*(x) + 16/116.)
    Ls = 116 * fscale(YbYn) - 16;       /* range [0, 100] */
    as = 500 * (fscale(XbXn) - fscale(YbYn));   /* range [-86.1813, 98.2352] */
    bs = 200 * (fscale(YbYn) - fscale(ZbZn));   /* range [-107.862, 94.4758] */

    nl = (Ls - 0) / (100 - 0) * 255.999999;
    na = (as + 86.1814) / (98.2353 + 86.1814) * 256;
    nb = (bs + 107.863) / (94.4759 + 107.863) * 256;

    if (nl < 0 || na < 0 || nb < 0 || nl >= 256 || na >= 256 || nb >= 256) {
        printf("%6g %6g %6g %d %d %d\n", Ls, as, bs, nl, na, nb);
        fatal("CIE L*a*b* conversion: vales out of range?");
    }

    n1 = nl;
    n2 = na;
    n3 = nb;
}


static void toCIELUV(colour_t r, colour_t g, colour_t b)
{
    double X, Y, Z, Ls, us, vs;
    int nl, nu, nv;

    X = 0.412453 * r + 0.357580 * g + 0.180423 * b;
    Y = 0.212671 * r + 0.715160 * g + 0.072169 * b + 0.0000001;
    Z = 0.019334 * r + 0.119193 * g + 0.950227 * b;

#define up(X,Y) (4*(X)/((X)+15*(Y)+3*(Z)))
#define vp(X,Y) (9*(Y)/((X)+15*(Y)+3*(Z)))

    Ls = 116 * fscale(Y / Yn) - 16;     /* range [0, 100] */
    us = 13 * Ls * (up(X, Y) - up(Xn, Yn));     /* range [-123.324, 147.573] */
    vs = 13 * Ls * (vp(X, Y) - vp(Xn, Yn));     /* range [-139.267, 12.1182] */

    nl = (Ls - 0) / (100 - 0) * 255.999999;
    nu = (us + 123.325) / (147.574 + 123.325) * 256;
    nv = (vs + 139.268) / (12.1183 + 139.268) * 256;
    if (nl < 0 || nu < 0 || nv < 0 || nl >= 256 || nu >= 256 || nv >= 256) {
        printf("%6g %6g %6g %d %d %d\n", Ls, us, vs, nl, nu, nv);
        fatal("CIE L*u*v* conversion: vales out of range?");
    }

    n1 = nl;
    n2 = nu;
    n3 = nv;
}


/*
 * HSV space (= cone) and HSL space (= double cone)
 *
 * H = Hue, S= saturation, V = brightness computed as max(r,g,b)
 *
 * As a cosequence, there are quite a few (r,g,b) pixels with V=1.
 * but only one with V=0, this is often visualised as a cone with the
 * apex symbolising V=0 and the base for V=1. Mathematically, HSV space
 * is a cylinder, though with H in [0,360) degrees, and L, S in [0,1].
 *
 * H = hue, S = saturation, L = brightness as (max(r,g,b)-min(r,g,b))/2
 *
 * There is one (r,g,b) pixel with L=0 and one with L=1. Most (r,g,b) pixels
 * mapped onto a particular L value will be for L=0.5. This is referred to
 * double-cone, with two apices at L=0 and L=1, respectively. Mathematically,
 * HSL space is a cylinder with H and S being radial coordinates and L
 * being the height.
 * 
 */

/*
 *
 * HSXcelllinear cuts up the volume of all possible H, S and X = L or V
 * values  in a linear way, ie, H space is linearly divided up in NH 
 * cells, S space is linearly divided up in NS cells and X space is linearly
 * cut up in NX cells. But note that all H cells for which S=0 are merged
 * into one bin.
 *
 * HSVcellvolume cuts up the volume of all possible H, S and V values into
 * the same number of cells as with above method, but seeks a non-linear
 * division of the S and V axis, so that the resuting bins cover all the
 * same number of (r,g,b) pixels of the whole RGB cube.
 *
 * HSLcellvolume cuts up the volume of all possible H, S and L values into
 * the same number of cells as with above method, but seeks a non-linear
 * division of the S and L axis, so that the resuting bins cover all the
 * same number of (r,g,b) pixels of the whole RGB cube.
 *
 */

static void toHSXlinear(double H, double S, double X)
{
    int h, s, x;

    h = H / 360 * 256;
    s = S * 255.999999;
    x = X * 255.999999;
    if (h < 0 || s < 0 || x < 0 || h >= 256 || s >= 256 || x >= 256) {
        printf("%6g %6g %6g %d %d %d\n", H, S, X, h, s, x);
        fatal("HSX linear conversion: vales out of range?");
    }

    /* conflate all hue cells where s=0 (grey, h undefined or nearly irrelevant) */

    n1 = h;
    n2 = s;
    n3 = x;
}

static void toHSVvolume(double H, double S, double V)
{
    int h, s, v;

    h = H / 360 * 256;
    s = (1 + (S * S * (256 * (256 - 1) + 1) - 1) / 256) * .99999999;
    v = V * V * 255.999999;
    if (h < 0 || s < 0 || v < 0 || h >= 256 || s >= 256 || v >= 256) {
        printf("%6g %6g %6g %d %d %d\n", H, S, V, h, s, v);
        fatal("HSV volume conversion: vales out of range?");
    }

    n1 = h;
    n2 = s;
    n3 = v;
}


static void toHSLvolume(double H, double S, double L)
{
    int h, s, l;

    h = H / 360 * 256;
    s = (1 + (S * S * (256 * (256 - 1) + 1) - 1) / 256) * .99999999;
    l = (L < .5 ? 2 * L * L : 1 - 2 * (1 - L) * (1 - L)) * 255.999999;
    if (h < 0 || s < 0 || l < 0 || h >= 256 || s >= 256 || l >= 256) {
        printf("%6g %6g %6g %d %d %d\n", H, S, L, h, s, l);
        fatal("HSL volume conversion: vales out of range?");
    }

    /* conflate all hue cells where s=0 (grey, h undefined or nearly irrelevant) */

    n1 = h;
    n2 = s;
    n3 = l;
}


static void toallHSV(colour_t r, colour_t g, colour_t b, void (*toHSX) (double H, double S, double V))
{
    int min, max;
    double delta, H, S, V;

    min = r;
    if (g < min)
        min = g;
    if (b < min)
        min = b;
    max = r;
    if (g > max)
        max = g;
    if (b > max)
        max = b;
    V = max / 255.;

    delta = max - min;
    S = delta ? delta / max : 0;

    H = -1;
    if (delta) {
        if (r == max)
            H = (g - b) / delta;        /* between yellow & magenta */
        else if (g == max)
            H = 2 + (b - r) / delta;    /* between cyan & yellow */
        else
            H = 4 + (r - g) / delta;    /* between magenta & cyan */

        H *= 60;                /* degrees */
        if (H < 0)
            H += 360;
    }

    toHSX(H, S, V);
}

static void toHSVlin(colour_t r, colour_t g, colour_t b)
{
    toallHSV(r, g, b, toHSXlinear);
}

static void toHSVvol(colour_t r, colour_t g, colour_t b)
{
    toallHSV(r, g, b, toHSVvolume);
}


/*
 * HSL space 
 */

static void toallHSL(colour_t r, colour_t g, colour_t b, void (*toHSX) (double H, double S, double L))
{
    int min, max;
    double delta, H, S, L;

    min = r;
    if (g < min)
        min = g;
    if (b < min)
        min = b;
    max = r;
    if (g > max)
        max = g;
    if (b > max)
        max = b;
    L = (max + min) / (255. * 2);

    delta = max - min;
    S = !delta ? 0 : L < 0.5 ? delta / (max + min) : delta / (2 * 255 - max - min);

    H = -1;
    if (delta) {
        if (r == max)
            H = (g - b) / delta;        /* between yellow & magenta */
        else if (g == max)
            H = 2 + (b - r) / delta;    /* between cyan & yellow */
        else
            H = 4 + (r - g) / delta;    /* between magenta & cyan */

        H *= 60;                /* degrees */
        if (H < 0)
            H += 360;
    }

    toHSX(H, S, L);
}

static void toHSLlin(colour_t r, colour_t g, colour_t b)
{
    toallHSL(r, g, b, toHSXlinear);
}

static void toHSLvol(colour_t r, colour_t g, colour_t b)
{
    toallHSL(r, g, b, toHSLvolume);
}


/*
 * converts an RGB image to Y'CbCr space
 */


AImage RGBtoYpCbCr(AImage im)
{
    unsigned char *c;
    int i, j, r, g, b;

    for (c = im.cont, j = 0; j < im.h; j++)
        for (i = 0; i < im.w; i++, c += 3) {
            r = c[0], g = c[1], b = c[2];
            toYpCbCr(r, g, b);  // sets n1, n2, n3
            c[0] = n1, c[1] = n2, c[2] = n3;
        }

    return im;
}
