/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 *
 * readjpg.c
 *
 * Allocates an image in memory, reads the jpeg image into this
 * space and returns the image.
 *
 * Authors
 * 
 * Joao Magalhaes, 12 April 2005
 *    - Image rescaling function modified to generate images with a 
 *      constant number of 8x8 blocks.
 * 
 * Stefan Rueger, Mar 2002
 * skeleton taken from Marcus Pickering, February 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <jpeglib.h>

#define DSTATE_READY    202     /* found SOS, ready for start_decompress */

#include "annotate.h"

#define RED 0.3
#define GREEN 0.59
#define BLUE 0.11

AImage ReadImageFile(char *filename)
{
    static struct jpeg_decompress_struct cinfo;
    static struct jpeg_error_mgr jerr;
    FILE *infile;
    int stride, i, n, bw=0;
    JSAMPARRAY line;
    AImage ret;

    if ((infile = fopen(filename, "r")) == NULL)
        fatal("failed to open %s", filename);

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);

    jpeg_stdio_src(&cinfo, infile);

    /* set the output colour space to YCbCr if required
     * is not always supported
     *  if(pars.useYCbCr)
     *    cinfo.out_color_space = JCS_YCbCr;
     */

    jpeg_read_header(&cinfo, TRUE);
    jpeg_start_decompress(&cinfo);

    ret.w = cinfo.output_width;
    ret.h = cinfo.output_height;
    switch (cinfo.output_components) {
    case 3:
        bw = 0;
        break;
    case 1:
        bw = 1;
        break;
    default:
        fatal("cannot deal with %d output components (should be 1 or 3)", cinfo.output_components);
    }
    /* ret.c = 3; always produce colour image */
    ret.t = colour;
    ret.cont = clrmalloc(ret.w * ret.h * 3 * sizeof *ret.cont);
    stride = cinfo.output_width * cinfo.output_components;
    line = (*cinfo.mem->alloc_sarray) ((j_common_ptr) & cinfo, JPOOL_IMAGE, stride, 1);

    for (n = 0; cinfo.output_scanline < cinfo.output_height;) {
        jpeg_read_scanlines(&cinfo, line, 1);
        if (bw)
            for (i = 0; i < stride; i++)
                ret.cont[n] = ret.cont[n + 1] = ret.cont[n + 2] = (*line)[i], n += 3;
        else
            for (i = 0; i < stride; i++)
                ret.cont[n++] = (*line)[i];
    }

    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);

    fclose(infile);
    ret.h = pars.maskbottom >= ret.h ? 1 : ret.h - pars.maskbottom;

    return ret;
}

AImage sampleDCT(AImage im, int nBlocks)
{
    AImage ret = im;
    int nrow, ncol, imPos, retPos;
    float scale, row, col, wError, hError, rowScale, colScale;

	int	blockSize = 64;    
 
    if (im.h <= 0 || im.w <= 0 || nBlocks < 10 || (scale = sqrt(((double) im.h * im.w / (nBlocks*blockSize) ))) < 1)
        return im;
	  
	ret.w = im.w/scale; 
    ret.h = im.h/scale; 

	if (pars.flag.dct) {
		hError = ret.h % 8;
		wError = ret.w % 8;
		if ((wError*ret.h) < (hError*ret.w)) {
			ret.w = (ret.w/8) * 8;
			ret.h = ((nBlocks) / (ret.w/8))*8;
		} else {
			ret.h = (ret.h/8) * 8;
			ret.w = ((nBlocks) / (ret.h/8))*8;
		}
		
		rowScale = (float)im.h/ret.h;
		colScale = (float)im.w/ret.w;
	} else {
		rowScale = scale;
		colScale = scale;
	}

	if ((rowScale < 1) || (colScale < 1))
		return im;
		
    ret.cont = clrmalloc(ret.w * ret.h * 3 * sizeof *ret.cont);
     
	for (row = 0, nrow=0; nrow < ret.h; row+=rowScale, nrow++)
		for (col = 0, ncol=0; ncol < ret.w; col+=colScale, ncol++) {
			imPos = (((int)row)*im.w + (int)col) * 3;
			retPos = (nrow*ret.w + ncol) * 3;
			ret.cont[retPos] = im.cont[imPos];
			ret.cont[retPos+1] = im.cont[imPos+1];
			ret.cont[retPos+2] = im.cont[imPos+2];
		}
   
    efree(im.cont);
    return ret;
}


/*
 * returns an image of roughly linear size k (say 256) by sampling
 * every d-th row/column - the original image is freed
 */
AImage sample(AImage im, int k) {
  AImage ret = im; 
  int d, j, ni, nj, dp;
  JSAMPLE *p, *q;

  if(im.h<=0 || im.w<=0 || k<10 || (d = sqrt(((double)im.h*im.w/k/k))+.5) < 2)
    return im;

  ret.w = (im.w+d-1)/d, ret.h = (im.h+d-1)/d;
  ret.cont = clrmalloc(ret.w * ret.h * 3 * sizeof*ret.cont);
  for(dp = (d-1)*3, q=ret.cont, j=nj=0; nj<ret.h; nj++, j+=d)
    for(p= &im.cont[im.w*3*j], ni=0; ni<ret.w; ni++, p+=dp)
      *q++ = *p++, *q++ = *p++, *q++ = *p++;
  efree(im.cont);
  //  printf("downsampled %d\n", d);
  return ret;
}



/*
 * converts the content of an image to greyscale - quantization is specified by
 * quant. The type is set to bw.
 */
AImage ConvertToGrey(AImage im, int quant)
{
    AImage ret = im;
    unsigned char *c;
    int i, j;
    double factor;
    ret.t = bw;
    ret.q = quant;
    ret.h = im.h;
    ret.w = im.w;
    factor = quant / 256.0;
    ret.cont = clrmalloc(im.w * im.h + sizeof *ret.cont);
    if (im.w > 0 && im.h > 0) {
        c = im.cont;
        for (i = 0; i < im.h; i++) {
            for (j = 0; j < im.w; j++, c += 3) {
                ret.cont[i * im.w + j] = (unsigned char) (factor * (c[0] * RED + c[1] * GREEN + c[2] * BLUE));
            }
        }
    }
    return ret;
}

void WriteImageFile(AImage im, FILE * outfile)
{
    static struct jpeg_compress_struct cinfo;
    static struct jpeg_error_mgr jerr;
    int stride;
    JSAMPROW row[1];

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    jpeg_stdio_dest(&cinfo, outfile);

    cinfo.image_width = im.w;
    cinfo.image_height = im.h;
    cinfo.input_components = 3; /* im.c; */
    cinfo.in_color_space = JCS_RGB;     /* im.c==3? JCS_RGB: JCS_GRAYSCALE; */

    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, 100, TRUE);

    jpeg_start_compress(&cinfo, TRUE);

    stride = im.w * 3;          /* im.c; */

    while (cinfo.next_scanline < cinfo.image_height) {
        row[0] = im.cont + cinfo.next_scanline * stride;
        jpeg_write_scanlines(&cinfo, row, 1);
    }

    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);
}

int cleanup;

static void sighandler(int sig)
{
    if (cleanup && pars.ftfname)
        unlink(pars.ftfname);
    fatal("caught signal %d, terminating...", sig);
}

static void catchsignals(int on)
{
    cleanup = on;
    signal(SIGHUP, on ? sighandler : SIG_DFL);
    signal(SIGINT, on ? sighandler : SIG_DFL);
    signal(SIGQUIT, on ? sighandler : SIG_DFL);
    signal(SIGILL, on ? sighandler : SIG_DFL);
    signal(SIGABRT, on ? sighandler : SIG_DFL);
    signal(SIGFPE, on ? sighandler : SIG_DFL);
    signal(SIGSEGV, on ? sighandler : SIG_DFL);
    signal(SIGPIPE, on ? sighandler : SIG_DFL);
    signal(SIGALRM, on ? sighandler : SIG_DFL);
    signal(SIGTERM, on ? sighandler : SIG_DFL);
}

FILE *ftfopen(char *ftfname)
{
    FILE *fp = NULL;
    int fd;

    catchsignals(1);
    if ((fd = open(ftfname, (pars.flag.f ? O_TRUNC : O_EXCL) | O_CREAT | O_WRONLY, 0666)) != -1) {
        if ((fp = fdopen(fd, "w")) == NULL)
            fatal("cannot fdopen %s in mode \"w\"", ftfname);
    } else {
        if (errno != EEXIST)
            warning("unable to create %s (continuing)", ftfname);
        errno = 0;
    }

    if (!fp)
        catchsignals(0);

    return fp;
}

void ftfclose(FILE * fp)
{
    if (fp) {
        if (fclose(fp))
            fatal("error when closing file %s", pars.ftfname ? pars.ftfname : "???");
        catchsignals(0);
        efree(pars.ftfname);
    }
}
