/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * smoothing.c
 *
 * smoothes colour pixels with a Gaussian kernel
 * two variants smoothing_G(r,g,b) and smoothing_S(r,g,b)
 * _G uses a fixed grid in coluor space and _S a relative grid to (r,g,b)
 *
 * Stefan M Rueger
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>

#include "annotate.h"

/*
 * smoothing_G(r,g,b)
 *
 * will be invoked on annotation modifier ":G<gridsz>-<nblocks>-<sigma>"
 * The smoothing function returns a list of N coordinates with 
 * relative weights (which sum up to sum) of how a single RGB colour
 * spreads in colour space to gridpoints (the grid points are <gridsz>
 * pixels apart per axis). The model is Gaussian with standard deviation
 * sigma. Not all gridpoints are updated: only nblocks around the (r,g,b)
 * per axis are updated.
 *
 */

Coord *smoothing_G(int r, int g, int b) {
  int n1, n2, n3, n1ll, n2ll, n3ll, n1i, n2i, n3i, d1, d2, d3, d11, d22, d33;
  int ci, ca, done, size;
  static int GN, NB, NBh, GNmNB, NC;
  static Coord *coor, **cache, head, *space;
  static double otss, *expval;

  if(!GN) {
    GN = 256/pars.gridG, NB = pars.nblocks, NBh = NB/2, GNmNB = GN-NB, NC = NB*(256+GN-1)/GN;
    otss = 1/(2.*(pars.sigma? pars.sigma: .1)*(pars.sigma? pars.sigma: .1));
    if(!expval && pars.memleft > (size = 3*NC*NC*sizeof*expval)) {
      expval = clrmalloc(size), pars.memleft -= size;
      for(ci=0; ci<3*NC*NC; ci++)
        expval[ci] = -1;
    }
    if(!cache && !pars.flag.G && pars.memleft > (size = NC*NC*NC*sizeof*cache))
      cache = clrmalloc(size), pars.memleft -= size;
  }

  n1=r*GN/256, n2=g*GN/256, n3=b*GN/256; // map to grid
 // lower left down neighbour grid coords
  n1ll=n1-NBh; if(n1ll<0) n1ll=0; else if(n1ll>GNmNB) n1ll=GNmNB;
  n2ll=n2-NBh; if(n2ll<0) n2ll=0; else if(n2ll>GNmNB) n2ll=GNmNB;
  n3ll=n3-NBh; if(n3ll<0) n3ll=0; else if(n3ll>GNmNB) n3ll=GNmNB;
  head.r = r, head.g=g, head.b=b; // head.index = (n1ll*GN+n2ll)*GN+n3ll;
  r -= n1ll*256/GN, g -= n2ll*256/GN, b -= n3ll*256/GN;

  if(r<0 || g<0 || b<0 || r>= NC || g>=NC || b>=NC) {
    fprintf(stderr, "r=%d, g=%d, b=%d, n1ll=%d, n2ll=%d, n3ll=%d\n", head.r, head.g, head.b, n1ll, n2ll, n3ll);
    fatal("how come r=%d, g=%d, b=%d out of range (%d)?", r, g, b, NC);
  }

  if(cache && cache[ca = (r*NC+g)*NC+b])
    done = 1, coor = cache[ca];
  else if(cache && pars.memleft > (size = sizeof**cache + NB*NB*NB*sizeof cache[0]->c[0]))
    done = 0, coor = cache[ca] = clrmalloc(size), pars.memleft-=size;
  else
    done = 0, coor = space? space: (space = clrmalloc(sizeof**cache + NB*NB*NB*sizeof cache[0]->c[0]));

  if(!done) {
    ci = 0, coor->sum = 0;
    for(n1i=0; n1i<NB; n1i++)
    for(d1 = (256*n1i+128)/GN - r, d11=d1*d1, n2i=0; n2i<NB; n2i++)
    for(d2 = (256*n2i+128)/GN - g, d22=d11+d2*d2, n3i=0; n3i<NB; n3i++) {
      coor->c[ci].dr = d1;
      coor->c[ci].dg = d2; 
      coor->c[ci].db = d3 = (256*n3i+128)/GN - b;
      if((d33 = d22+d3*d3) >= 3*NC*NC)
        fatal("d33 (%d) out of range [0, %d]?", d33, 3*NC*NC);
      if(expval[d33]<0)
        expval[d33] = exp(-d33*otss);
      coor->sum += coor->c[ci].w = expval[d33];
//      coor->c[ci].dindex = (n1i*GN+n2i)*GN+n3i;
      ci++;
    }
    coor->N=NB*NB*NB;
    if(coor->sum < 1e-16)
      ci = ((n1-n1ll)*NB+(n2-n2ll))*NB+(n3-n3ll), coor->c[ci].w = 1.0;
    else
      for(ci=0; ci<coor->N; ci++)
        coor->c[ci].w /= coor->sum;
    coor->sum = 1.0;
  }
  coor->r = head.r, coor->g = head.g, coor->b = head.b; // coor->index = head.index; 

  return coor;
}


/*
 * smoothing_S(r,g,b)
 *
 * will be invoked on ":S<gridsz>-<nblocks>-<sigma>" 
 * The smoothing function returns a list of N coordinates with 
 * relative weights (which sum up to sum) of how a single RGB colour
 * spreads in colour space to gridpoints relative to (r,g,b).
 * The stepsize per axis is <gridsz> and the nblocks**3 closest
 * ones are updated. The model is Gaussian with standard deviation sigma.
 *
 */

Coord *smoothing_S(int r, int g, int b) {
  int nx, n1l, n2l, n3l, n1u, n2u, n3u, n1s, n2s, n3s, d1, d2, d3, d1s, d2s, d3s, d11, d22, d33;
  int ci, ca, done, size;
  static int GS, NB, NBh, CAOFF;
  static Coord *coor, **cache, *space;
  static double otss, *expval;

  if(!GS) {
    GS = pars.gridS, NB = pars.nblocks, NBh=NB/2, CAOFF = ((NB-1)*NB+NB-1)*NB+NB-1;
    otss = GS*GS/(2.*(pars.sigma? pars.sigma: .1)*(pars.sigma? pars.sigma: .1));
    if(!expval && pars.memleft > (size = 3*NB*NB*sizeof*expval)) {
      expval = clrmalloc(size), pars.memleft -= size;
      for(ci=0; ci<3*NB*NB; ci++)
        expval[ci] = -1;
    }
    if(!cache && !pars.flag.G && pars.memleft > (size = NB*NB*NB*sizeof*cache))
      cache = clrmalloc(size), pars.memleft -= size;
  }

  nx=r/GS; n1l=nx<NBh? -nx: -NBh; if((nx=(r+(n1l+NB)*GS-256)/GS)>0) n1l-=nx;
  nx=g/GS; n2l=nx<NBh? -nx: -NBh; if((nx=(g+(n2l+NB)*GS-256)/GS)>0) n2l-=nx;
  nx=b/GS; n3l=nx<NBh? -nx: -NBh; if((nx=(b+(n3l+NB)*GS-256)/GS)>0) n3l-=nx;

  n1u=n1l+NB-1; n1s=n1l*GS;
  n2u=n2l+NB-1; n2s=n2l*GS;
  n3u=n3l+NB-1; n3s=n3l*GS;


//  ca = (n1l*NB+n2l)*NB+n3l + CAOFF;
//  if(n1l<=-NB || n2l<=-NB || n3l<=-NB || n1l>0 || n2l>0 || n3l>0 || ca<0 || ca>=NB*NB*NB)
//    fatal("how come n1l=%d, n2l=%d or n3l=%d out of bound [%d,0]", n1u, n2u, n3u, -NB+1);

  ca = (n1u*NB+n2u)*NB+n3u;
  if(n1u<0 || n2u<0 || n3u<0 || n1u>=NB || n2u>=NB || n3u>=NB)
    fatal("how come n1u=%d, n2u=%d or n3u=%d out of bound [0,%d]", n1u, n2u, n3u, NB);

  if(cache && cache[ca])
    done = 1, coor = cache[ca];
  else if(cache && pars.memleft > (size = sizeof**cache + NB*NB*NB*sizeof cache[0]->c[0]))
    done = 0, coor = cache[ca] = clrmalloc(size), pars.memleft-=size;
  else
    done = 0, coor = space? space: (space = clrmalloc(sizeof**cache + NB*NB*NB*sizeof cache[0]->c[0]));

  if(!done) {
    ci = 0, coor->sum = 0;
    for(d1=n1l, d1s=n1s; d1<=n1u; d1++, d1s+=GS)
    for(d11=d1*d1, d2=n2l, d2s=n2s; d2<=n2u; d2++, d2s+=GS)
    for(d22=d11+d2*d2, d3=n3l, d3s=n3s; d3<=n3u; d3++, d3s+=GS) {
      coor->c[ci].dr = d1s;
      coor->c[ci].dg = d2s; 
      coor->c[ci].db = d3s;
      if((d33 = d22+d3*d3) >= 3*NB*NB)
        fatal("d33 (%d) out of range [0, %d]?", d33, 3*NB*NB);
      if(expval[d33]<0)
        expval[d33] = exp(-d33*otss);
      coor->sum += coor->c[ci].w = expval[d33];
      ci++;
    }
    coor->N=NB*NB*NB;
    if(coor->sum < 1e-16)
      fatal("sum should not be zero (%f)", coor->sum);
    for(ci=0; ci<coor->N; ci++)
      coor->c[ci].w /= coor->sum;
    coor->sum = 1.0;
  }
  coor->r = r, coor->g = g, coor->b = b, coor->index = 0; 

  return coor;
}


/*********
int gauss(FILE *fp, Image im) {
  int i, j, k, GN=256/pars.gridG;
  unsigned char *c;
  typedef struct { int index; double weight; } Coords;
  Coord *co;
  static float *hist;
  if(!hist)
    hist = clrmalloc(GN*GN*GN*sizeof*hist);
  else
    memset(hist, 0, GN*GN*GN*sizeof*hist);

  if(im.w>0 && im.h>0) {
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) {
        co = smoothing_G(c[0], c[1], c[2]);
        if(co->sum)
          for(k=0; k<co->N; k++)
            hist[co->index+co->c[k].dindex] += co->c[k].w;
      }
  }

  fwrite(hist, sizeof*hist, GN*GN*GN, fp);
  printf("memleft %d\n", pars.memleft/1024/1024);
  return 1;
}
********/
