/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * tamura.c
 *
 * Peter Howarth
 * November 2003
 * 
 * Calculates 3 of the tamura texture features for an image - coarseness,
 * contrast and directionality
 *
 * tamuraS - calculates the 3 features for the entire image
 *
 * tamuraImage - generates and saves a jpg where red=coarseness, green=contrast
 * and blue=directionality. images of each separate channel can be made
 * 
 * tamuraCND - this uses the rgb functionality from colourhist.c to build a 3d histogram
 * from the tamura image
 *
 * margTamura - this uses margcolouhist.c to generate marginal color hsitograms or central
 * moments of the tamura image 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"


#define max(a,b) ((a>b) ? a : b)
#define DEBUG pars.flag.d
#define THRESHOLD 12
#define NO_OF_INTERVALS 16
#define CK_FACTOR 0.9

//static int NO_OF_INTERVALS = 16;

static double RED = 0.3, GREEN = 0.59, BLUE = 0.11;
static float CONTRAST_MAX = 60;
static float COARSENESS_MAX;    // in fact this depends on 2^k
static float DIRECTION_MAX = 2.0;
static float D_MIN = 0.0;

static float entropy(double hist[NO_OF_INTERVALS]);

typedef struct {
  int h, w, k;
  int *image;
}Ak;

typedef struct {
  int h, w;
  int *image;
}Ad;

static void calculateMoments(int **g_image,
			     int row, int col, 
			     int winH, int winW, 
			     double *mean, double *variance,
			     double *skewness, double *kurtosis)
{
  int i, j;
  double area = (double)winH * winW;
  (*mean)=(*variance)=(*skewness)=(*kurtosis)=0;
  for (i=row; i < row + winH; i++) {
    for (j=col; j < col + winW; j++) {
      (*mean)+=(double)(g_image[i][j]);
    }
  }
  (*mean) /= area;
   
  for (i=row; i < row + winH; i++) {
    for (j=col; j < col + winW; j++) {
      double diff = (double)(g_image[i][j] - *mean);
      double mult = diff;
      *variance += (mult*=diff);
      *skewness += (mult*=diff);
      *kurtosis += (mult*=diff);
    }
  }
  *variance /= area;
  *skewness /= area;
  *kurtosis /= area;
}


static int quantizeAngle(double angle, int intervals)
{
  angle  = (angle * intervals / M_PI) + 0.5;
  return (int)fmod(angle , (double)intervals);
}


static Ad setupAd(int **g_image, int width, int height) {
  int i, j, dh, dw, mod_g;
  double theta;
  Ad ret;
  ret.h = height - 2;
  ret.w = width - 2;
  // set up the array for ak
  if (ret.h>0 && ret.w>0) {
    ret.image = clrmalloc(ret.h * ret.w *sizeof(*ret.image));
    // fill up the array
    for (i = 1; i < height - 1; i++) {
      for (j = 1; j < width - 1; j++) {
	dw = g_image[i+1][j-1] + g_image[i+1][j] + g_image[i+1][j-1] - g_image[i-1][j-1] - g_image[i-1][j] - g_image[i-1][j+1];
	dh = g_image[i-1][j-1] + g_image[i][j-1] + g_image[i+1][j-1] - g_image[i-1][j+1] - g_image[i][j+1] - g_image[i+1][j+1];

	mod_g = (int)(fabs(dh) + fabs(dw))/2;
	if (mod_g > THRESHOLD) {
	  theta  = atan(((double)dw)/((double)dh)) + M_PI_2;
	  ret.image[(i-1)*ret.w + (j-1)] = quantizeAngle(theta, NO_OF_INTERVALS);
	} else {
	ret.image[(i-1)*ret.w + (j-1)] = -1;
	}
      } 
    }
  }
  return ret;
}

static void freeAd(Ad ad) {
  free(ad.image);
}

static void definePeak(double hist[NO_OF_INTERVALS], int* peak, int* left, int* right) {
  int i, level;
  int max = 0;
  *peak = 0;
  for(i=0; i<NO_OF_INTERVALS; i++) {
    level = hist[i];
    if (level > max) {
      max = level;
      *peak = i;
    }
  }
  // now find the valleys on either side
  i = *peak + 1;
  while ((hist[i%NO_OF_INTERVALS] > hist[(i+1)%NO_OF_INTERVALS]) && hist[i%NO_OF_INTERVALS] > 0.0) {
    i++;
  }
  *right = i % NO_OF_INTERVALS;

  i = *peak + NO_OF_INTERVALS - 1;
  while ((hist[i%NO_OF_INTERVALS] > hist[(i-1)%NO_OF_INTERVALS]) && hist[i%NO_OF_INTERVALS] > 0.0) {
    i--;
  }
  *left = i % NO_OF_INTERVALS;
  return;
}

static float tamDirection(double hist[NO_OF_INTERVALS], int left, int right, int peak) {
  double feature = 0.0;
  int i, width;
  if (DEBUG) printf("RAW: left=%d peak=%d right=%d\n", left, peak, right);
  width = (right > left) ? right - left : right - left + NO_OF_INTERVALS;
  peak = (peak > left) ? peak : peak + NO_OF_INTERVALS;
  for (i = left; i <= left + width; i++) {
    feature +=  (double)hist[i%NO_OF_INTERVALS] * pow((i-peak), 2);
  }
  return (float)feature;
}

static float directionWindow(Ad ad, int row, int col, int win) {
   int i, j, bin, total;
  int peak, left, right;
  float weight, dir;
   double hist[NO_OF_INTERVALS];
  for (i = 0; i < NO_OF_INTERVALS; i++)
    hist[i] = 0.0;
  total = 0;
  for (i=row; i < row + win; i++) {
    for (j=col; j < col + win; j++) {
      bin = ad.image[i*ad.w + j];
      if (bin >= 0) {
	hist[bin]++;
	total++;
      }
    }
  }
  definePeak(hist, &peak, &left, &right);
  // normalize the histogram
  for (i = 0; i < NO_OF_INTERVALS; i++) {
    if (DEBUG) printf("%4.1f ", hist[i]);
    hist[i] /= total;
  }
  if(DEBUG)printf("\n");
  if (total==0)
    return DIRECTION_MAX;
  // weight the result depending on the number of pixels above the threshold
  weight = 1.0;//(float)total / (float)(win * win);
  dir = tamDirection(hist, left, right, peak);
  return (dir/ weight);

} 

static float directionAllEntropy(Ad ad) {
  int i, j, bin, total;
  double hist[NO_OF_INTERVALS];
  for (i = 0; i < NO_OF_INTERVALS; i++)
    hist[i] = 0.0;
  total = 0;
  for (i=0; i < ad.h; i++) {
    for (j=0; j < ad.w; j++) {
      bin = ad.image[i*ad.w + j];
      if (bin >= 0) {
	hist[bin]++;
	total++;
      }
    }
  }
  if (total != 0) {
    for (i = 0; i < NO_OF_INTERVALS; i++) {
      if (DEBUG) printf("%4.1f ", hist[i]);
      hist[i] /= total;
    }
    if(DEBUG)printf("\n");
  }
  return entropy(hist);
} 



static float entropy(double hist[NO_OF_INTERVALS]) {
  int i;
  double ent = 0;
  for (i=0; i < NO_OF_INTERVALS; i++) {
    if (hist[i] != 0)
      ent -= (hist[i] * log(hist[i]) / log(2));
  } 
  return (float) ent;
} 

static float contrastWindow(int **g_image, int width, int height, int row, int col, int winH, int winW) 
{
  double mean, variance, skewness, kurtosis;
  calculateMoments(g_image, row, col, winH, winW, &mean, &variance, &skewness, &kurtosis);
  if (DEBUG) {
    printf("sd=%f kurtosis=%f\n", sqrt(variance), (kurtosis/(variance * variance)));
    printf("width=%d height=%d winH=%d winW=%d\n", width, height, winH,winW);
  }
  if (!variance || !kurtosis)
    return 0.0;
  return (float)(sqrt(variance) / pow((kurtosis / (variance*variance)), 0.25));
}

static int akIndex(Ak ak, int i, int j, int k) {
  return ((i * ak.w * ak.k) + (j * ak.k) + k);
}

static Ak setupAk(int **g_image, int width, int height, int kMax) {  
int i, j, ii, k, ip, jp, radius, bound;
  Ak ret;
  ret.image = NULL;
  bound = (int)pow(2, kMax-1);
  ret.h = height - (bound * 2);
  ret.w = width - (bound * 2);
  ret.k = kMax;

  if (ret.h>0 && ret.w>0 && ret.k>0) {
    ret.image = (int *)clrmalloc(ret.h * ret.w * ret.k * sizeof(*ret.image));

    for (i= bound; i < height - bound; i++) {
      for (j=bound; j < width - bound; j++) {
	for (ii = (i-bound)*ret.w*ret.k + (j-bound)*ret.k, k = 1; k <= ret.k; k++) {
	  radius = 1<<(k-1);
	  for (ip = i-radius; ip < i + radius; ip++) {
	    for (jp = j-radius; jp < j + radius; jp++) {
	      ret.image[ii + k -1] += g_image[ip][jp];
	    }
	  }
	  ret.image[ii + k -1] /= pow(2, 2*k);
	}
      }
    }
  }
  return  ret;
}

static void freeAk(Ak ak) {
  free(ak.image);
}


static int** createGreyscaleArray(AImage im) {
  int i, j;
  unsigned char *c;
  int **g_image = (int **)clrmalloc(im.h * sizeof(int *));
  for (i = 0; i < im.h; i++) {
    g_image[i] = (int *)clrmalloc(im.w * sizeof(int));
  }
  if(im.w>0 && im.h>0) {
    c=im.cont;
    for(i=0; i<im.h; i++) {
      for(j=0; j<im.w; j++, c+=3) {
	g_image[i][j]  = (int)(c[0]*RED + c[1]*GREEN + c[2]*BLUE);
      }
    }
  }
  return g_image;
}

static int getBin(float measure, int noOfBins, float range) {
  int bin;
  bin = (int)(measure * noOfBins / range);
  if (bin <= 0)
    return 0;
  return  (bin < noOfBins) ? bin : (noOfBins - 1);
}

static float coarseAll(Ak ak) {
  int i, j, k, bound, radius;
  char kHat;
  float coarseness = 0.0;
  double ck, ckMax;
  bound = pow(2, (ak.k - 1));
  for(i = bound; i < ak.h - bound ;i++) {
    for(j = bound; j < ak.w - bound ;j++) {
      ckMax = 0.0;
      kHat = 0;
      for (k=1; k <= ak.k ; k++) {
	radius = 1<<(k-1);
	ck =  max(fabs(ak.image[akIndex(ak, i-radius, j, k-1)] - ak.image[akIndex(ak, i+radius, j, k-1)]),
			fabs(ak.image[akIndex(ak, i, j-radius, k-1)] - ak.image[akIndex(ak, i, j+radius, k-1)]));
	if (ck > CK_FACTOR * ckMax) {
	  kHat = k;
	  if (ck > ckMax)
	    ckMax = ck;
	}
      }
      coarseness += 1<<kHat;
    }
  }
  coarseness /= (ak.h - 2*bound)*(ak.w-2*bound);
  return coarseness;
}

static float coarseWindow(Ak ak, int row, int col, int win) {
  int i, j, k, bound, radius;
  float coarseness = 0.0;
  double ckMax, ck;
  char kHat;
  // register image row and column to ak row and column by subtracting bound
  bound = pow(2, (ak.k - 1));
  for(i = row - bound; i < row + win - bound;i++) {
    for(j = col - bound; j < col + win -bound ;j++) {
      ckMax = 0.0;
      kHat = 0;
      for (k=1; k <= ak.k ; k++) {
	radius = 1<<(k-1);
	ck =  max(fabs(ak.image[akIndex(ak, i-radius, j, k-1)] - ak.image[akIndex(ak, i+radius, j, k-1)]),
			fabs(ak.image[akIndex(ak, i, j - radius, k-1)] - ak.image[akIndex(ak, i, j+radius, k-1)]));
	if (ck > CK_FACTOR * ckMax) {
	  kHat = k;
	  if (ck > ckMax)
	    ckMax = ck;
	}
      }
      // ** use kHat rather than 2**kHat - improves retrival performance
      coarseness += kHat;
    }	
  }
  coarseness /= (win)*(win);
  return coarseness;
}

/* This function calculates the tamura features: contrast, coarseness and directionality for
 * the entire image and saves a vector of the features. It is ideal for use with simple
 * tiling
 * NOTE: for small images the k value for coarseness may be reduced
 */
extern int tamuraS(FILE *fp, AImage im) {
  int i, k, size, maxk, cont, coarse, dir, n;
  int **g_image = createGreyscaleArray(im);
  float *result;
  Ak ak; 
  Ad ad;

  k = pars.k;
  coarse = pars.flag.t[0];
  cont =  pars.flag.t[1];
  dir =  pars.flag.t[2];
  result = clrmalloc((coarse + cont + dir) * sizeof*result);
  n = 0;


  if (coarse) {
    // reduce the value of k if the borders would be too big for the image
    size = im.h < im.w ? im.h/2 : im.w/2;
    maxk = (int)floor(log(size-1)/log(2));
    k = k > maxk ? maxk : k;
    ak = setupAk(g_image, im.w, im.h, k);
    result[n] = coarseAll(ak);
    if (DEBUG)
      printf("k=%d coarseness=%f\n", k, result[n]);
    n++;
    freeAk(ak);
  }     
  if (cont) {
    result[n] = contrastWindow(g_image, im.w, im.h, 0, 0, im.h, im.w);
    if (DEBUG)
      printf("contrast=%f\n", result[n]);
    n++;
  }
  
  if (dir) {
    ad = setupAd(g_image, im.w, im.h);
    result[n] = directionAllEntropy(ad);
    if (DEBUG)
      printf("direction=%f\n", result[n]);
    n++;
    freeAd(ad);
  }  
  fwrite(result, sizeof*result, n, fp); 
  for(i = 0; i < im.h; i++) {
    free(g_image[i]);
  }
  efree(result);
  free(g_image);
  return 1;
}

int tamuraImageCheck() {
  pars.win = pars.number[0];
  if (pars.win < 1) {
    fatal("Window size should be greater than 0 in tamuraImage-<W>-<K>-[ACND].  Default value is 8");
  }
  pars.k = pars.number[1];
  if (pars.k < 0) {
    fatal("K value should be 0 or positive in tamuraImage-<W>-<K>-[ACND].  Default value is 6");
  }
  switch(*pars.space[0]) {
  case 'A': case 'a':  //pars.ftfext = "_tam.Jpg"; 
    pars.flag.t[0]=1; pars.flag.t[1]=1; pars.flag.t[2]=1; break;
  case 'C': case 'c': //pars.ftfext =  "_tam_coarseness.Jpg";
     pars.flag.t[0]=1; pars.flag.t[1]=0; pars.flag.t[2]=0; break;
  case 'N': case 'n': //pars.ftfext = "_tam_contrast.Jpg"; 
    pars.flag.t[0]=0; pars.flag.t[1]=1; pars.flag.t[2]=0; break;
  case 'D': case 'd': //pars.ftfext = "_tam_direction.Jpg";
    pars.flag.t[0]=0; pars.flag.t[1]=0; pars.flag.t[2]=1; break;
  default:
    fatal("must specify A, C, N, or D in tamuraImage-<W>-<K>-[ACND].  Default is A=all. C=coarseness, N=contrast, D=directionality");
  }
  pars.ftfext = ".Jpg";
  return 1;
}

int tamuraImage(FILE *fp, AImage im) {
  AImage t_im = convertToTamura(im, pars.flag.t[0], pars.flag.t[1], pars.flag.t[2]);
  WriteImageFile(t_im, fp);
  free(t_im.cont);
  return 1;
}

static int setTFlags(void) {
  int i;
  pars.useTamura = 1;
  for(i=0; i<3; i++) {
    if(pars.number[i]<=1)
      pars.flag.t[i] = 0;
    else
      pars.flag.t[i] = 1;
  }
  pars.win = pars.number[3];
  if (pars.win < 1) {
    fatal("Window size = %d, it should be greater than 0 in ...-[GFSLC]-<W>-<K>. Default value is 8", pars.win);
  }
  pars.k = pars.number[4];
  if (pars.k < 0) {
    fatal("K value should be 0 or positive in ...-[GFSLC]-<W>-<K>.  Default value is 6");
  }
  return 1;
}

int tamuraCheck() {
  pars.k = pars.number[0];
  if (pars.k<0) {
        fatal("must specify a value of k >= 0, default value is 6\n");
  }

 switch(*pars.space[0]) {
  case 'A': case 'a': pars.flag.t[0]=1; pars.flag.t[1]=1; pars.flag.t[2]=1; break;
  case 'C': case 'c': pars.flag.t[0]=1; pars.flag.t[1]=0; pars.flag.t[2]=0; break;
  case 'N': case 'n': pars.flag.t[0]=0; pars.flag.t[1]=1; pars.flag.t[2]=0; break;
  case 'D': case 'd': pars.flag.t[0]=0; pars.flag.t[1]=0; pars.flag.t[2]=1; break;
  default:
    fatal("must specify A, C, N, or D in tamuraS-<K>-[ACND].  Default is A=all. C=coarseness, N=contrast, D=directionality");
  }
  return 1;
}

int histTamuraCheck() {
  XYZcheck();
  setTFlags();
  return 1;
}

int margTamuraCheck() {
  margXYZcheck();
  setTFlags();
  return 1;
}


AImage convertToTamura(AImage im, int coarse, int cont, int dir) {
  int i, j, n, kMax, bound, win, BINS, binN, binC, binD;
  float contrast, directionality, coarseness;
  int **g_image = createGreyscaleArray(im);
  AImage t_im;
  Ak ak;
  Ad ad;

  win = pars.win;
  BINS = 256;
  kMax = pars.k;
  bound = pow(2, kMax);
  COARSENESS_MAX = kMax; // the k value sets the maximum range of coarseness
 
  t_im.h = im.h - win - 2*bound;
  t_im.w = im.w - win - 2*bound;
  if (t_im.h < 1 || t_im.w < 1) {
    fatal("the combined window and k values are too large for this image -- try with k <= 6\n");    
  }
  t_im.cont = clrmalloc((t_im.h) * (t_im.w) * 3 * sizeof*t_im.cont);
  n = 0;
  i = bound;
  if (coarse)
    ak = setupAk(g_image, im.w, im.h, kMax);
  if (dir)
    ad = setupAd(g_image, im.w, im.h);
  while (i < t_im.h + bound) {
    j = bound;
    while (j < t_im.w + bound) {
      if (cont) {
	contrast= contrastWindow(g_image, im.w, im.h, i, j, win, win);
	binN = (unsigned int)getBin(contrast, BINS, CONTRAST_MAX);
      } else {
	binN = 0;
      }
      if (coarse) {
	coarseness = coarseWindow(ak, i, j, win);
	binC = (unsigned int)getBin(coarseness, BINS, COARSENESS_MAX);
      } else {
	binC = 0;
      }
      if (dir) {
	directionality = directionWindow(ad, i, j, win);
	binD = 255 - (unsigned int)getBin((directionality-D_MIN), BINS, DIRECTION_MAX-D_MIN);
      } else {
	binD = 0;
      }
      t_im.cont[n] = binC;
      t_im.cont[n+1] = binN;
      t_im.cont[n+2] = binD;
      n += 3;
      j += 1;
    }
    i += 1;
  }

  for (i = 0; i < im.h; i++)
    free(g_image[i]);
  free(g_image);
  if(coarse)
    freeAk(ak);
  if(dir)
    freeAd(ad);
  return t_im;
}
